#!/usr/bin/env python
#
# Copyright (c) 2009-2022 CERN. All rights nots expressly granted are
# reserved.
#
# This file is part of iLCDirac
# (see ilcdirac.cern.ch, contact: ilcdirac-support@cern.ch).
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# In applying this licence, CERN does not waive the privileges and
# immunities granted to it by virtue of its status as an
# Intergovernmental Organization or submit itself to any jurisdiction.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.
#
"""Test the Combined Software Installation class."""

from __future__ import absolute_import
import unittest
import os
from mock import patch, MagicMock as Mock

import pytest
import six

from DIRAC import S_OK, S_ERROR
from ILCDIRAC.Core.Utilities.CombinedSoftwareInstallation import CombinedSoftwareInstallation, \
    getSharedAreaLocation, createSharedArea, getLocalAreaLocation, getSoftwareFolder, \
    getEnvironmentScript, checkCVMFS, extractTarball
from Tests.Utilities.GeneralUtils import assertEqualsImproved, assertDiracFailsWith, \
    assertDiracSucceeds, assertDiracSucceedsWith_equals, assertMockCalls

__RCSID__ = "$Id$"


class TestCombinedSWInstallation(unittest.TestCase):
  """Test the different methods of the class."""
  STD_DICT = {'Job': {'SoftwarePackages': 'mypackagev1.0',
                         'SystemConfig': 'mytestconfig',
                         'Platform': 'mytestplatform'},
               'CE': {'CompatiblePlatforms': 'blabla'}, 'Source': {}}

  def setUp(self):
    with patch('%s.NativeMachine.CMTSupportedConfig' % MODULE_NAME, new=Mock(return_value=['x86_64-slc5-gcc43-opt'])):
      self.csi = CombinedSoftwareInstallation({})
    self.csi.apps = [('myprogram', 'v6765')]

  def test_constructor(self):
    localapps = [('dep1', 'v4'), ['dep2', 'v5.0']]
    localapps_raw = ['dep1.v4', 'dep2.v5.0']
    self.csi = CombinedSoftwareInstallation({
        'Job': {'SoftwarePackages': localapps_raw,
                 'Platform': 'coolplatform123'}})
    assertEqualsImproved(self.csi.jobConfig, 'coolplatform123', self)
    assertEqualsImproved(self.csi.apps, localapps, self)

  def test_constructor_illegal_sw_package(self):
    import copy
    custom_job_dict = copy.deepcopy(TestCombinedSWInstallation.STD_DICT['Job'])
    custom_dict = copy.deepcopy(TestCombinedSWInstallation.STD_DICT)
    custom_job_dict['SoftwarePackages'] = 1
    del custom_job_dict['SystemConfig']
    custom_dict['Job'] = custom_job_dict
    custom_dict['CE'] = {'CompatiblePlatforms': ['iamcompatibletoo', 'here']}
    with patch('%s.NativeMachine.CMTSupportedConfig' % MODULE_NAME, new=Mock(return_value=['x86_64-slc5-gcc43-opt'])):
      self.csi = CombinedSoftwareInstallation(custom_dict)
    assertEqualsImproved(self.csi.apps, [], self)
    assertEqualsImproved(self.csi.ceConfigs, ['x86_64-slc5-gcc43-opt'], self)

  def test_execute_simple(self):
    self.csi.apps = None
    result = self.csi.execute()
    assertDiracSucceedsWith_equals(result, None, self)
    assertEqualsImproved(self.csi.jobConfig, 'x86_64-slc5-gcc43-opt', self)

  def test_execute(self):
    with patch('%s.resolveDeps' % MODULE_NAME, new=Mock(return_value=[{'app': 'dependency123', 'version': '3.4'}])), \
        patch('%s.Operations.getSections' % MODULE_NAME, new=Mock(return_value=S_OK(['x86_64-slc5-gcc43-opt']))), \
        patch('%s.installInAnyArea' % MODULE_NAME, new=Mock(return_value=S_OK())), \
        patch('%s.NativeMachine.CMTSupportedConfig' % MODULE_NAME, new=Mock(return_value=['x86_64-slc5-gcc43-opt'])):
      self.csi = CombinedSoftwareInstallation(TestCombinedSWInstallation.STD_DICT)
      assertEqualsImproved(self.csi.ceConfigs, ['x86_64-slc5-gcc43-opt'], self)
      assertEqualsImproved(self.csi.jobConfig, 'mytestconfig', self)
      result = self.csi.execute()
      assertDiracSucceeds(result, self)
      assertEqualsImproved(self.csi.jobConfig, 'x86_64-slc5-gcc43-opt', self)

  def test_execute_noconfig(self):
    self.csi.jobConfig = None
    assertDiracFailsWith(self.csi.execute(), 'no architecture requested', self)

  def test_execute_getsections_fails(self):
    with patch('%s.Operations.getSections' % MODULE_NAME, new=Mock(return_value=S_ERROR('getsection_err'))):
      assertDiracFailsWith(self.csi.execute(), 'getsection_err', self)

  def test_execute_not_compatible(self):
    with patch('%s.resolveDeps' % MODULE_NAME, new=Mock(return_value=[{'app': 'dependency123', 'version': '3.4'}])), \
        patch('%s.Operations.getSections' % MODULE_NAME, new=Mock(return_value=S_OK(['some_Exotic_system1', 'nope_not_this_one_either']))), \
        patch('%s.installInAnyArea' % MODULE_NAME, new=Mock(return_value=S_OK())), \
        patch('%s.NativeMachine.CMTSupportedConfig' % MODULE_NAME, new=Mock(return_value=['x86_64-slc5-gcc43-opt'])):
      import copy
      custom_dict = copy.deepcopy(TestCombinedSWInstallation.STD_DICT)
      custom_dict['CE'] = {'CompatiblePlatforms': ['iamcompatibletoo',
                                                      'here']}
      self.csi = CombinedSoftwareInstallation(custom_dict)
      result = self.csi.execute()
      assertDiracFailsWith(result, 'requested architecture not supported by ce', self)

  def test_execute_locally(self):
    with patch('%s.resolveDeps' % MODULE_NAME, new=Mock(return_value=[{'app': 'dependency123', 'version': '3.4'}])), \
        patch('%s.Operations.getSections' % MODULE_NAME, new=Mock(return_value=S_OK([]))), \
        patch('%s.installInAnyArea' % MODULE_NAME, new=Mock(return_value=S_OK())), \
        patch('%s.checkCVMFS' % MODULE_NAME, new=Mock(side_effect=[S_OK(), S_ERROR()])), \
        patch('%s.installInAnyArea' % MODULE_NAME, new=Mock(side_effect=[S_OK(), S_OK()])), \
        patch('%s.createSharedArea' % MODULE_NAME, new=Mock(return_value=True)):
      import copy
      self.csi = CombinedSoftwareInstallation(TestCombinedSWInstallation.STD_DICT)
      self.csi.ceConfigs = []
      self.csi.sharedArea = ''
      result = self.csi.execute()
      assertDiracSucceeds(result, self)

  def test_execute_installfails(self):
    with patch('%s.resolveDeps' % MODULE_NAME, new=Mock(return_value=[{'app': 'dependency123', 'version': '3.4'}])), \
        patch('%s.Operations.getSections' % MODULE_NAME, new=Mock(return_value=S_OK([]))), \
        patch('%s.installInAnyArea' % MODULE_NAME, new=Mock(return_value=S_OK())), \
        patch('%s.checkCVMFS' % MODULE_NAME, new=Mock(side_effect=[S_OK(), S_ERROR()])), \
        patch('%s.installInAnyArea' % MODULE_NAME, new=Mock(return_value=S_ERROR('could not install my test program'))), \
        patch('%s.createSharedArea' % MODULE_NAME, new=Mock(return_value=True)):
      import copy
      self.csi = CombinedSoftwareInstallation(TestCombinedSWInstallation.STD_DICT)
      self.csi.ceConfigs = []
      self.csi.sharedArea = ''
      result = self.csi.execute()
      assertDiracFailsWith(result, 'could not install my test program', self)

  def test_execute_install_dependency_fails(self):
    with patch('%s.resolveDeps' % MODULE_NAME, new=Mock(return_value=[{'app': 'dependency123', 'version': '3.4'}])), \
        patch('%s.Operations.getSections' % MODULE_NAME, new=Mock(return_value=S_OK([]))), \
        patch('%s.installDependencies' % MODULE_NAME, new=Mock(side_effect=[S_OK(), S_ERROR()])), \
        patch('%s.createSharedArea' % MODULE_NAME, new=Mock(return_value=False)), \
        patch('%s.installInAnyArea' % MODULE_NAME, new=Mock(return_value=S_OK())):
      import copy
      self.csi = CombinedSoftwareInstallation(TestCombinedSWInstallation.STD_DICT)
      self.csi.ceConfigs = []
      self.csi.sharedArea = ''
      result = self.csi.execute()
      assertDiracFailsWith(result, 'failed to install dep', self)

  def test_execute_nosharedarea(self):
    with patch('%s.resolveDeps' % MODULE_NAME, new=Mock(return_value=[{'app': 'dependency123', 'version': '3.4'}])), \
        patch('%s.Operations.getSections' % MODULE_NAME, new=Mock(return_value=S_OK([]))), \
        patch('%s.getSharedAreaLocation' % MODULE_NAME, new=Mock(return_value='')), \
        patch('%s.createSharedArea' % MODULE_NAME, new=Mock(return_value=True)), \
        patch('%s.installInAnyArea' % MODULE_NAME, new=Mock(return_value=S_OK())):
      import copy
      self.csi = CombinedSoftwareInstallation(TestCombinedSWInstallation.STD_DICT)
      self.csi.ceConfigs = []
      self.csi.sharedArea = ''
      result = self.csi.execute()
      assertDiracSucceeds(result, self)

  def test_listareadir_nofail(self):
    with patch('%s.systemCall' % MODULE_NAME, new=Mock(return_value=S_OK([0, 'important_message', 'my_subprocess_error_msg']))), \
        patch('%s.LOG.info' % MODULE_NAME,
              new=Mock(side_effect=[True, KeyError('injecting this into logger call')])) as mock_log:
      from ILCDIRAC.Core.Utilities.CombinedSoftwareInstallation import listAreaDirectory
      try:
        listAreaDirectory(self.csi.sharedArea)
        self.fail('Should not reach this due to KeyError being thrown')
      except KeyError as ke:
        mock_log.assert_any_call('important_message')
        assertEqualsImproved(ke.__repr__(),
                             ("KeyError('injecting this into logger call')" if six.PY3 else
                              "KeyError('injecting this into logger call',)"),
                              self)

  def test_listareadir_fail_a_bit(self):
    with patch('%s.systemCall' % MODULE_NAME, new=Mock(return_value=S_OK([1, 'entry', 'my_subprocess_error_msg']))), \
        patch('%s.LOG.error' % MODULE_NAME,
              new=Mock(side_effect=KeyError('injecting this into logger call 2'))) as mock_log:
      from ILCDIRAC.Core.Utilities.CombinedSoftwareInstallation import listAreaDirectory
      try:
        listAreaDirectory(self.csi.sharedArea)
        self.fail('Should not reach this due to KeyError being thrown')
      except KeyError as ke:
        mock_log.assert_called_with('Failed to list the area directory',
                                     'my_subprocess_error_msg')
        assertEqualsImproved(ke.__repr__(),
                             ("KeyError('injecting this into logger call 2')" if six.PY3 else
                              "KeyError('injecting this into logger call 2',)"),
                              self)

  def test_listareadir_fail_completely(self):
    with patch('%s.systemCall' % MODULE_NAME, new=Mock(return_value=S_ERROR('some_os_listdir_error'))), \
        patch('%s.LOG.error' % MODULE_NAME,
              new=Mock(side_effect=KeyError('injecting this into logger call 3'))) as mock_log:
      from ILCDIRAC.Core.Utilities.CombinedSoftwareInstallation import listAreaDirectory
      try:
        listAreaDirectory(self.csi.sharedArea)
        self.fail('Should not reach this due to KeyError being thrown')
      except KeyError as ke:
        mock_log.assert_called_with('Failed to list the area directory',
                                     'some_os_listdir_error')
        assertEqualsImproved(ke.__repr__(),
                             ("KeyError('injecting this into logger call 3')" if six.PY3 else
                              "KeyError('injecting this into logger call 3',)"),
                              self)


class TestSharedLocation(unittest.TestCase):
  """Tests the sharedArea and localArea functions."""

  def test_getsharedarealoc(self):
    exists_dict = {'mylocation123test': True}
    with patch('%s.Operations.getValue' % MODULE_NAME, new=Mock(return_value='mylocation123test')), \
         patch('%s.os.path.exists' % MODULE_NAME, new=Mock(side_effect=lambda path: exists_dict[path])), \
         patch('%s.DIRAC.gConfig.getValue' % MODULE_NAME, new=Mock(side_effect=['a', 'a', '', ''])), \
         patch('%s.os.path.isdir' % MODULE_NAME, new=Mock(return_value=True)):
      result = getSharedAreaLocation()
      assertEqualsImproved(result, 'mylocation123test', self)

  def test_getsharedarealoc_environvar(self):
    exists_dict = {'testLocation135': False, '/abc/def/ghi/clic': True}
    with patch('%s.Operations.getValue' % MODULE_NAME, new=Mock(return_value=['testLocation135', '$MANY_MORE_LOCATIONS'])), \
         patch.dict(os.environ, {'MANY_MORE_LOCATIONS': '/abc/def/ghi'}, True), \
         patch('%s.os.path.exists' % MODULE_NAME, new=Mock(side_effect=lambda path: exists_dict[path])), \
         patch('%s.DIRAC.gConfig.getValue' % MODULE_NAME, new=Mock(side_effect=['a', 'a', '', ''])), \
         patch('%s.os.path.isdir' % MODULE_NAME, new=Mock(return_value=True)):
      result = getSharedAreaLocation()
      assertEqualsImproved(result, '/abc/def/ghi/clic', self)

  def test_getsharedarealoc_environvar_notfound(self):
    exists_dict = {'testLocation135': False, '/abc/def/ghi/clic': False}
    with patch('%s.Operations.getValue' % MODULE_NAME, new=Mock(return_value=['testLocation135', '$I_AM_FAKE', '$MANY_MORE_LOCATIONS'])), \
         patch.dict(os.environ, {'MANY_MORE_LOCATIONS': '/abc/def/ghi'}, True), \
         patch('%s.os.path.exists' % MODULE_NAME, new=Mock(side_effect=lambda path: exists_dict[path])), \
         patch('%s.DIRAC.gConfig.getValue' % MODULE_NAME, new=Mock(side_effect=['a', 'a', '', ''])), \
         patch('%s.os.path.isdir' % MODULE_NAME, new=Mock(return_value=True)):
      result = getSharedAreaLocation()
      assertEqualsImproved(result, '/abc/def/ghi/clic', self)

  def test_getsharedarealoc_overwrite_via_config(self):
    exists_dict = {'testLocation135': False, '/abc/def/ghi/clic': True}
    with patch('%s.Operations.getValue' % MODULE_NAME, new=Mock(return_value=['testLocation135', '$MANY_MORE_LOCATIONS'])), \
         patch.dict(os.environ, {'MANY_MORE_LOCATIONS': '/abc/def/ghi'}, True), \
         patch('%s.os.path.exists' % MODULE_NAME, new=Mock(side_effect=lambda path: exists_dict[path])), \
         patch('%s.DIRAC.gConfig.getValue' % MODULE_NAME, new=Mock(side_effect=['a', 'a', '/myotherpath/hereissharedarea', '/myotherpath/hereissharedarea'])), \
         patch('%s.os.path.isdir' % MODULE_NAME, new=Mock(return_value=True)):
      result = getSharedAreaLocation()
      assertEqualsImproved(result, '/myotherpath/hereissharedarea', self)

  def test_getsharedarealoc_notadir(self):
    exists_dict = {'testLocation135': False, '/abc/def/ghi/clic': True}
    with patch('%s.Operations.getValue' % MODULE_NAME, new=Mock(return_value=['testLocation135', '$MANY_MORE_LOCATIONS'])), \
         patch.dict(os.environ, {'MANY_MORE_LOCATIONS': '/abc/def/ghi'}, True), \
         patch('%s.os.path.exists' % MODULE_NAME, new=Mock(side_effect=lambda path: exists_dict[path])), \
         patch('%s.DIRAC.gConfig.getValue' % MODULE_NAME, new=Mock(side_effect=['a', 'a', '', ''])), \
         patch('%s.os.path.isdir' % MODULE_NAME, new=Mock(return_value=False)):
      result = getSharedAreaLocation()
      assertEqualsImproved(result, '', self)

  def test_getsharedarealoc_notfound(self):
    with patch('%s.Operations.getValue' % MODULE_NAME, new=Mock(return_value=[])), \
        patch('%s.os.path.exists' % MODULE_NAME) as exists_mock, \
        patch('%s.DIRAC.gConfig.getValue' % MODULE_NAME, new=Mock(side_effect=['a', 'a', '', ''])), \
        patch('%s.os.path.isdir' % MODULE_NAME, new=Mock(return_value=True)):
      result = getSharedAreaLocation()
      assertEqualsImproved(result, '', self)
      self.assertFalse(exists_mock.called)

  def test_createsharedarea(self):
    with patch.dict(os.environ, {'VO_ILC_SW_DIR': '/myilc/sharedarea/cooldir', 'OSG_APP': '/appdir/shared'}, True), \
        patch('%s.os.path.isdir' % MODULE_NAME, new=Mock(return_value=True)) as isdir_mock, \
        patch('%s.os.path.islink' % MODULE_NAME, new=Mock(return_value=False)) as islink_mock, \
        patch('%s.os.remove' % MODULE_NAME, new=Mock(return_value=True)) as remove_mock, \
        patch('%s.os.path.exists' % MODULE_NAME, new=Mock(return_value=False)) as exists_mock, \
        patch('%s.os.makedirs' % MODULE_NAME, new=Mock(return_value=True)) as mkdirs_mock:
      result = createSharedArea()
      self.assertTrue(result)
      self.assertFalse(remove_mock.called)
      self.assertFalse(exists_mock.called)
      self.assertFalse(mkdirs_mock.called)
      finalsharedarea = '/myilc/sharedarea/cooldir/clic'
      isdir_mock.assert_called_once_with(finalsharedarea)
      islink_mock.assert_called_once_with(finalsharedarea)

  def test_createsharedarea_noenvvars(self):
    with patch.dict(os.environ, {}, True), \
            patch('%s.LOG.info' % MODULE_NAME) as mock_log:
      result = createSharedArea()
      self.assertFalse(result)
      mock_log.assert_called_with('VO_ILC_SW_DIR and OSG_APP not defined.', )

  def test_createsharedarea_points_to_curdir(self):
    with patch.dict(os.environ, {'OSG_APP': '.'}, True), \
            patch('%s.LOG.info' % MODULE_NAME) as mock_log:
      result = createSharedArea()
      self.assertFalse(result)
      mock_log.assert_called_with('VO_ILC_SW_DIR or OSG_APP points to "."', )

  def test_createsharedarea_makedirs(self):
    with patch.dict(os.environ, {'VO_ILC_SW_DIR': '/myilc/sharedarea/cooldir', 'OSG_APP': '/appdir/shared'}, True), \
        patch('%s.os.path.isdir' % MODULE_NAME, new=Mock(return_value=True)) as isdir_mock, \
        patch('%s.os.path.islink' % MODULE_NAME, new=Mock(return_value=True)) as islink_mock, \
        patch('%s.os.remove' % MODULE_NAME, new=Mock(return_value=True)) as remove_mock, \
        patch('%s.os.path.exists' % MODULE_NAME, new=Mock(return_value=False)) as exists_mock, \
        patch('%s.os.makedirs' % MODULE_NAME, new=Mock(return_value=True)) as mkdirs_mock:
      result = createSharedArea()
      self.assertTrue(result)
      self.assertFalse(remove_mock.called)
      finalsharedarea = '/myilc/sharedarea/cooldir/clic'
      isdir_mock.assert_called_once_with(finalsharedarea)
      islink_mock.assert_called_once_with(finalsharedarea)
      exists_mock.assert_called_once_with(finalsharedarea)
      mkdirs_mock.assert_called_once_with(finalsharedarea)

  def test_createsharedarea_remove_and_recreate(self):
    with patch.dict(os.environ, {'OSG_APP': '/appdir/shared'}, True), \
        patch('%s.os.path.isdir' % MODULE_NAME, new=Mock(return_value=False)) as isdir_mock, \
        patch('%s.os.path.islink' % MODULE_NAME, new=Mock(return_value=False)) as islink_mock, \
        patch('%s.os.remove' % MODULE_NAME, new=Mock(return_value=True)) as remove_mock, \
        patch('%s.os.path.exists' % MODULE_NAME, new=Mock(return_value=True)) as exists_mock, \
        patch('%s.os.makedirs' % MODULE_NAME, new=Mock(return_value=True)) as mkdirs_mock:
      result = createSharedArea()
      self.assertTrue(result)
      self.assertFalse(islink_mock.called)
      finalsharedarea = '/appdir/shared/clic'
      isdir_mock.assert_called_once_with(finalsharedarea)
      exists_mock.assert_called_once_with(finalsharedarea)
      mkdirs_mock.assert_called_once_with(finalsharedarea)
      remove_mock.assert_called_once_with(finalsharedarea)

  def test_createsharedarea_oserr(self):
    with patch.dict(os.environ, {'VO_ILC_SW_DIR': '/myilc/sharedarea/cooldir', 'OSG_APP': '/appdir/shared'}, True), \
        patch('%s.os.path.isdir' % MODULE_NAME, new=Mock(return_value=True)) as isdir_mock, \
        patch('%s.os.path.islink' % MODULE_NAME, new=Mock(return_value=True)) as islink_mock, \
        patch('%s.os.remove' % MODULE_NAME, new=Mock(side_effect=OSError('some_filesys_error'))) as remove_mock, \
        patch('%s.os.path.exists' % MODULE_NAME, new=Mock(return_value=True)) as exists_mock, \
        patch('%s.os.makedirs' % MODULE_NAME, new=Mock(return_value=True)) as mkdirs_mock, \
            patch('%s.LOG.error' % MODULE_NAME) as mock_err:
      result = createSharedArea()
      self.assertFalse(result)
      self.assertFalse(mkdirs_mock.called)
      finalsharedarea = '/myilc/sharedarea/cooldir/clic'
      isdir_mock.assert_called_once_with(finalsharedarea)
      islink_mock.assert_called_once_with(finalsharedarea)
      exists_mock.assert_called_once_with(finalsharedarea)
      remove_mock.assert_called_once_with(finalsharedarea)
      mock_err.assert_called_with('Problem trying to create shared area',
                                   'some_filesys_error')

  def test_getlocalarealoc_already_exists(self):
    with patch('%s.DIRAC.gConfig.getValue' % MODULE_NAME, new=Mock(return_value='/gconfig/localarea')), \
        patch('%s.os.path.isdir' % MODULE_NAME, new=Mock(return_value=True)) as isdir_mock, \
        patch('%s.os.path.exists' % MODULE_NAME, new=Mock(return_value=False)) as exists_mock, \
        patch('%s.os.remove' % MODULE_NAME, new=Mock(return_value=True)) as remove_mock, \
        patch('%s.os.mkdir' % MODULE_NAME, new=Mock(return_value=True)) as mkdir_mock:
      result = getLocalAreaLocation()
      isdir_mock.assert_called_once_with('/gconfig/localarea')
      self.assertFalse(remove_mock.called)
      self.assertFalse(mkdir_mock.called)
      self.assertFalse(exists_mock.called)
      assertEqualsImproved(result, '/gconfig/localarea', self)

  def test_getlocalarealoc_remove_old(self):
    import DIRAC
    oldRoot = DIRAC.rootPath
    DIRAC.rootPath = '/dirac/myrootpath'
    with patch('%s.DIRAC.gConfig.getValue' % MODULE_NAME, new=Mock(return_value='')), \
         patch('%s.os.path.isdir' % MODULE_NAME, new=Mock(return_value=False)) as isdir_mock, \
         patch('%s.os.path.exists' % MODULE_NAME, new=Mock(return_value=True)) as exists_mock, \
         patch('%s.os.remove' % MODULE_NAME, new=Mock(return_value=True)) as remove_mock:
      result = getLocalAreaLocation()
      finallocalarea = '/dirac/myrootpath/LocalArea'
      isdir_mock.assert_called_once_with(finallocalarea)
      exists_mock.assert_called_once_with(finallocalarea)
      remove_mock.assert_called_once_with(finallocalarea)
      assertEqualsImproved(result, finallocalarea, self)
    DIRAC.rootPath = oldRoot

  def test_getlocalarealoc_create_new_dir(self):
    with patch('%s.DIRAC.gConfig.getValue' % MODULE_NAME, new=Mock(return_value='/gconfig/localarea')), \
        patch('%s.os.path.isdir' % MODULE_NAME, new=Mock(return_value=False)) as isdir_mock, \
        patch('%s.os.path.exists' % MODULE_NAME, new=Mock(return_value=False)) as exists_mock, \
        patch('%s.os.remove' % MODULE_NAME, new=Mock(return_value=True)) as remove_mock, \
        patch('%s.os.mkdir' % MODULE_NAME, new=Mock(return_value=True)) as mkdir_mock:
      result = getLocalAreaLocation()
      finallocalarea = '/gconfig/localarea'
      isdir_mock.assert_called_once_with(finallocalarea)
      mkdir_mock.assert_called_once_with(finallocalarea)
      exists_mock.assert_called_once_with(finallocalarea)
      self.assertFalse(remove_mock.called)
      assertEqualsImproved(result, finallocalarea, self)

  def test_getlocalarealoc_remove_fails(self):
    with patch('%s.DIRAC.gConfig.getValue' % MODULE_NAME, new=Mock(return_value='/gconfig/localarea')), \
        patch('%s.os.path.isdir' % MODULE_NAME, new=Mock(return_value=False)) as isdir_mock, \
        patch('%s.os.path.exists' % MODULE_NAME, new=Mock(return_value=True)) as exists_mock, \
        patch('%s.os.remove' % MODULE_NAME, new=Mock(side_effect=OSError('some_remove_oserror'))) as remove_mock, \
        patch('%s.os.mkdir' % MODULE_NAME, new=Mock(return_value=True)) as mkdir_mock, \
            patch('%s.LOG.error' % MODULE_NAME) as mock_err:
      result = getLocalAreaLocation()
      finallocalarea = '/gconfig/localarea'
      isdir_mock.assert_called_once_with(finallocalarea)
      exists_mock.assert_called_once_with(finallocalarea)
      remove_mock.assert_called_once_with(finallocalarea)
      mock_err.assert_called_with('Cannot remove:', '/gconfig/localarea because some_remove_oserror')
      self.assertFalse(mkdir_mock.called)
      assertEqualsImproved(result, '', self)

  def test_getlocalarealoc_create_fails(self):
    with patch('%s.DIRAC.gConfig.getValue' % MODULE_NAME, new=Mock(return_value='/gconfig/localarea')), \
        patch('%s.os.path.isdir' % MODULE_NAME, new=Mock(return_value=False)) as isdir_mock, \
        patch('%s.os.path.exists' % MODULE_NAME, new=Mock(return_value=False)) as exists_mock, \
        patch('%s.os.remove' % MODULE_NAME, new=Mock(return_value=True)) as remove_mock, \
        patch('%s.os.mkdir' % MODULE_NAME, new=Mock(side_effect=OSError('some_mkdir_oserror'))) as mkdir_mock, \
            patch('%s.LOG.error' % MODULE_NAME) as mock_err:
      result = getLocalAreaLocation()
      finallocalarea = '/gconfig/localarea'
      isdir_mock.assert_called_once_with(finallocalarea)
      exists_mock.assert_called_once_with(finallocalarea)
      mkdir_mock.assert_called_once_with(finallocalarea)
      mock_err.assert_called_with('Cannot create:', '/gconfig/localarea because some_mkdir_oserror')
      self.assertFalse(remove_mock.called)
      assertEqualsImproved(result, '', self)

  def test_getsoftwarefolder(self):
    exists_dict = {'myapparchive.test.tgz': False,
                   '/mylocalarea/test/myapparchive.test': False,
                   '/testshared/area/myapparchive.test': True}
    with patch('%s.Operations.getValue' % MODULE_NAME, new=Mock(return_value='myapparchive.test.tgz')) as getval_mock, \
         patch('%s.getLocalAreaLocation' % MODULE_NAME, new=Mock(return_value='/mylocalarea/test')), \
         patch('%s.getSharedAreaLocation' % MODULE_NAME, new=Mock(return_value='/testshared/area')), \
         patch('%s.os.path.exists' % MODULE_NAME, new=Mock(side_effect=lambda path: exists_dict[path])) as exists_mock:
      result = getSoftwareFolder('a', 'b', 'c')
      assertMockCalls(exists_mock, ['/mylocalarea/test/myapparchive.test',
                                      '/testshared/area/myapparchive.test', 'myapparchive.test.tgz'], self)
      getval_mock.assert_called_with('/AvailableTarBalls/a/b/c/TarBall', '')
      assertDiracSucceedsWith_equals(result, '/testshared/area/myapparchive.test', self)

  def test_getsoftwarefolder_from_cvmfs(self):
    with patch('%s.checkCVMFS' % MODULE_NAME, new=Mock(return_value=S_OK(('mycvmfsfolder/txt', 'otherentry')))) as cvmfs_mock:
      result = getSoftwareFolder('a', 'b', 'c')
      cvmfs_mock.assert_called_once_with('a', ['b', 'c'])
      assertDiracSucceedsWith_equals(result, 'mycvmfsfolder/txt', self)

  def test_getsoftwarefolder_apptar_fails(self):
    with patch('%s.Operations.getValue' % MODULE_NAME, new=Mock(return_value='')):
      result = getSoftwareFolder('a', 'b', 'c')
      assertDiracFailsWith(result, 'could not find b, c name from cs', self)

  def test_getsoftwarefolder_uselocal(self):
    exists_dict = {'myapparchivev2.test.tar.gz': False, '/mylocalarea/test/myapparchivev2.test': True}
    with patch('%s.Operations.getValue' % MODULE_NAME, new=Mock(return_value='myapparchivev2.test.tar.gz')) as getval_mock, \
         patch('%s.getLocalAreaLocation' % MODULE_NAME, new=Mock(return_value='/mylocalarea/test')), \
         patch('%s.getSharedAreaLocation' % MODULE_NAME, new=Mock(return_value='/testshared/area')), \
         patch('%s.os.path.exists' % MODULE_NAME, new=Mock(side_effect=lambda path: exists_dict[path])) as exists_mock:
      result = getSoftwareFolder('a', 'b', 'c')
      assertMockCalls(exists_mock, ['/mylocalarea/test/myapparchivev2.test', 'myapparchivev2.test.tar.gz'],
                       self)
      getval_mock.assert_called_with('/AvailableTarBalls/a/b/c/TarBall', '')
      assertDiracSucceedsWith_equals(result, '/mylocalarea/test/myapparchivev2.test', self)

  def test_getsoftwarefolder_notfound(self):
    exists_dict = {'myapp_executable': False,
                   '/mylocalarea/test/myapp_executable': False,
                   '/testshared/area/myapp_executable': False}
    with patch('%s.Operations.getValue' % MODULE_NAME, new=Mock(return_value='myapp_executable')) as getval_mock, \
         patch('%s.getLocalAreaLocation' % MODULE_NAME, new=Mock(return_value='/mylocalarea/test')), \
         patch('%s.getSharedAreaLocation' % MODULE_NAME, new=Mock(return_value='/testshared/area')), \
         patch('%s.os.path.exists' % MODULE_NAME, new=Mock(side_effect=lambda path: exists_dict[path])) as exists_mock:
      result = getSoftwareFolder('a', 'b', 'c')
      assertMockCalls(exists_mock, ['/mylocalarea/test/myapp_executable', '/testshared/area/myapp_executable',
                                      'myapp_executable'], self)
      getval_mock.assert_called_with('/AvailableTarBalls/a/b/c/TarBall', '')
      assertDiracFailsWith(result, 'missing installation of myapp_executable', self)

  def test_getEnvironmentScript(self):
    with patch('%s.checkCVMFS' % MODULE_NAME, new=Mock(return_value=S_OK(('/otherfolder/otherfile', '/cvmfsfolder/myenvscript')))):
      result = getEnvironmentScript('a', 'b', 'c', None)
      assertDiracSucceedsWith_equals(result, '/cvmfsfolder/myenvscript', self)

  def test_getEnvironmentScript_cvmfs_empty(self):
    def return_my_value(platform, appname, appversion):  # pylint: disable=missing-docstring
      return (platform, appname, appversion)
    with patch('%s.checkCVMFS' % MODULE_NAME, new=Mock(return_value=S_OK(('entry', '')))):
      result = getEnvironmentScript('a', 'b', 'c', return_my_value)
      assertEqualsImproved(result, ('a', 'b', 'c'), self)

  def test_getEnvironmentScript_from_passed_func(self):
    def return_my_value(platform, appname, appversion):  # pylint: disable=missing-docstring
      return (appname, platform, appversion)
    with patch('%s.checkCVMFS' % MODULE_NAME, new=Mock(return_value=S_ERROR('some_cvmfs_error'))):
      result = getEnvironmentScript('a', 'b', 'c', return_my_value)
      assertEqualsImproved(result, ('b', 'a', 'c'), self)

  def test_checkcvmfs_not_found(self):
    result = checkCVMFS('mytestplatform', ('appnametesttest123', 'appversionv77testp2'))
    assertDiracFailsWith(result, 'missing cvmfs', self)

  def test_checkcvmfs_exists(self):
    with patch('%s.Operations.getValue' % MODULE_NAME, new=Mock(side_effect=['testcvmfspath', 'testenvscript'])), \
        patch('%s.os.path.exists' % MODULE_NAME, new=Mock(return_value=True)):
      result = checkCVMFS('mytestplatform', ('appnametesttest123',
                                              'appversionv77testp2'))
      assertDiracSucceedsWith_equals(result, ('testcvmfspath',
                                               'testenvscript'), self)


CLASS_NAME = 'CombinedSoftwareInstallation'
MODULE_NAME = 'ILCDIRAC.Core.Utilities.%s' % CLASS_NAME


@pytest.mark.parametrize('success, filename, throw',
                         [(False, 'myfile', False),
                          (True, 'my.tar', False),
                          (True, 'my.tar.gz', False),
                          (False, 'noFile.tar.gz', True)])
def test_extractTarball(success, filename, throw, mocker):
  """Test the extractTarball function."""
  tarModule = Mock(name='TarModule')
  tarball = Mock(name='TarBall')
  tarModule.open.return_value = tarball
  if throw:
    tarModule.open.side_effect = RuntimeError('Nope')
  mocker.patch('%s.tarfile' % MODULE_NAME, tarModule)
  assert extractTarball(filename, 'dir')['OK'] == success
  if success:
    tarball.extractall.assert_called_once_with('dir')



#
# Copyright (c) 2009-2022 CERN. All rights nots expressly granted are
# reserved.
#
# This file is part of iLCDirac
# (see ilcdirac.cern.ch, contact: ilcdirac-support@cern.ch).
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# In applying this licence, CERN does not waive the privileges and
# immunities granted to it by virtue of its status as an
# Intergovernmental Organization or submit itself to any jurisdiction.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.
#
"""
Whizard2: Interface to the Whizard 2 generator

.. versionadded:: v26r0p14

Usage:

>>> whiz = Whizard2()
>>> whiz.setVersion("2.3.1")
>>> whiz.setNumberOfEvents(30)
>>> whiz.setRandomSeed(15)
>>> whiz.setSinFile("__path_to__/process.sin")

"""

from __future__ import absolute_import
import types
import os
import re

from ILCDIRAC.Interfaces.API.NewInterface.LCApplication import LCApplication
from DIRAC import S_OK, S_ERROR, gLogger
from DIRAC.Core.Workflow.Parameter import Parameter
from DIRAC.ConfigurationSystem.Client.Helpers.Operations import Operations
import six


LOG = gLogger.getSubLogger(__name__)

__RCSID__ = "$Id$"


class Whizard2(LCApplication):
  """Whizard2 Application Class."""

  def __init__(self, paramdict=None):
    self.randomSeed = -1
    self.eventType = ''
    self.whizard2SinFile = ''
    self.whizard2RawSin = False
    super(Whizard2, self).__init__(paramdict)
    # Those 5 need to come after default constructor
    self._modulename = 'Whizard2Analysis'
    self._moduledescription = 'Module to run Whizard2'
    self.appname = 'whizard2'
    self.datatype = 'GEN'
    self._paramsToExclude.extend(["outputDstPath", "outputRecPath", "OutputDstFile", "OutputRecFile"])
    self._ops = Operations()
    self._decayProc = ['decay_proc']
    self._integratedProcess = ''

  def setRandomSeed(self, randomSeed):
    """ Optional: Define random seed to use. Default is the jobID.

    :param int randomSeed: Seed to use during generation.
    """
    self._checkArgs({'randomSeed': int})
    self.randomSeed = randomSeed

  def setEvtType(self, evttype):
    """Define process. If the process given is not found, when calling :func:`UserJob.append() <ILCDIRAC.Interfaces.API.NewInterface.UserJob.UserJob.append>` a full list is printed.

    :param str evttype: Process to generate
    """
    self._checkArgs({'evttype': (str,)})
    if self.addedtojob:
      return self._reportError("Cannot modify this attribute once application has been added to Job")
    self.eventType = evttype
    return S_OK()

  def setProcessVariables(self, processes):
    """Set the list of processes to simulate.

    The sindarin file will be modified later on to call **simulate (proc_a, proc_b)**.
    The process variables have to be defined in the sindarin file::

      process proc_a ...
      process proc_b ...

    .. versionadded:: v28r0p6

    :param processes: which processes to call simulate on, by default 'decay_proc'
    :type processes: list, str
    """
    if isinstance(processes, six.string_types):
      self._decayProc = [proc.strip() for proc in processes.split(',')]
      return S_OK()
    elif isinstance(processes, (set, list, tuple)):
      self._decayProc = [proc.strip() for proc in processes]
      return S_OK()

    return self._reportError("Cannot handle this argument type")

  def setSinFile(self, whizard2SinFilePath, raw=False):
    """Set the Whizard2 options to be used.

    Usage:
      - Give path to the Whizard2 steering file.
      - IMPORTANT: set **seed** via iLCDirac API -> `Whizard2.setRandomSeed`
      - IMPORTANT: set **n_events** via iLCDirac API  -> `Whizard2.setNumberOfEvents`
      - IMPORTANT: set **OutputFile** via iLCDirac API -> `Whizard2.setOutputFile`
      - The variables in which processes are defined which should be simulated can be set
        via `Whizard2.setProcessVariables`

    :param str whizard2SinFilePath: Path to the whizard2 sin file.
    :param bool raw: If set to ``True`` only the **seed** will be added to the
      sindarin file. The **simulate(decay process)** , **n_events**, and **OutputFile** modifications by iLCDirac
      will be disabled. The user is responsible for consistency in this case.
    """
    self._checkArgs({'whizard2SinFilePath': (str,)})

    # Chech if file exist
    if not os.path.isfile(whizard2SinFilePath):
      return self._reportError('Whizard2 Sin file does not exist!')

    # Read file
    self.whizard2SinFile = open(whizard2SinFilePath).read()

    if raw:
      self.whizard2RawSin = True
      LOG.info("Using the raw sindarin file")
      return None

    if "n_events" in self.whizard2SinFile:
      return self._reportError('Do not set n_events in the sin file, set it via the iLCDirac API')

    if "seed" in self.whizard2SinFile:
      return self._reportError('Do not set seed in the sin file, set it via the iLCDirac API')

    if "simulate(" in self.whizard2SinFile.replace(" ", ""):
      return self._reportError('Do not call "simulate ()" in the sin file, this is done by iLCDirac')

    return None

  def setIntegratedProcess(self, integrationTarball):
    """Make whizard2 use an already integrated process.

    The ``integrationTarball`` has to be a tarball (zip, tar.gz, tgz), either an LFN, or a process defined in the
    configuration system.

    Use `getKnownProcesses` to see the list of defined processes

    >>> # processes defined in the configuration
    >>> whizard2.setIntegratedProcess('bbcbbc_3tev_negPol')

    >>> # tarball on the grid
    >>> whizard2.setIntegratedProcess('LFN:/ilc/user/u/username/bbcbbc_3tev_negPol.tar.gz')

    :param str integrationTarball: integrated process to be used for event generation

    .. warning :: It is the responsibility of the user to ensure that the sindarin file is compatible with
                  the integrated process

    .. note ::

      Whizard2 will be executed in the path where the tarball is extracted. Therefore, the tarball must contain all the
      files resulting from running Whizard2 directly and not in a subdirectory. The hidden ``.libs/`` folder must be
      included.

      To produce the tarball, change into the directory in which Whizard2 has been run and do::

        tar cvzf <name>.tar.gz * .libs

    .. note :: Tarballs stored on CVMFS need to be registered in the ConfigurationSystem.
    """
    self._checkArgs({'integrationTarball': (str,)})

    # file on the grid
    if integrationTarball.lower().startswith('lfn:'):
      LOG.info('Integrated process file is an LFN, adding it to the sandbox')
      self.inputSB.append(integrationTarball)
      # as the tarball is automatically extracted during the workflow, we do not have to do anything
      self._integratedProcess = ''
      return S_OK()

    knownProcesses = self.getKnownProcesses()
    if not knownProcesses['OK']:
      return self._reportError('Failed to get known integrated processes: %s' % knownProcesses['Message'])
    elif integrationTarball in knownProcesses['Value']:
      self._integratedProcess = integrationTarball
    else:
      return self._reportError('Unknown integrated process in %s: %s (available are: %s)' %
                               (self.appname, integrationTarball, ', '.join(list(knownProcesses['Value'].keys()))))
    return S_OK()

  def getKnownProcesses(self, version=None):
    """Return a list of known integrated processes.

    :param str version: Optional: Software version for which to print the integrated processes.
       If not given the version of the application instance is used.
    :returns: S_OK with list of integrated processes known for this software version, S_ERROR
    """
    if version is None and not self.version:
      return S_ERROR('No software version defined')
    version = self.version if version is None else version
    processes = self._ops.getOptionsDict('/AvailableTarBalls/%s/whizard2/%s/integrated_processes/processes' %
                                         ('x86_64-slc5-gcc43-opt', self.version))
    return processes

  def _userjobmodules(self, stepdefinition):
    res1 = self._setApplicationModuleAndParameters(stepdefinition)
    res2 = self._setUserJobFinalization(stepdefinition)
    if not res1["OK"] or not res2["OK"]:
      return S_ERROR('userjobmodules failed')
    return S_OK()

  def _prodjobmodules(self, stepdefinition):
    res1 = self._setApplicationModuleAndParameters(stepdefinition)
    res2 = self._setOutputComputeDataList(stepdefinition)
    if not res1["OK"] or not res2["OK"]:
      return S_ERROR('prodjobmodules failed')
    return S_OK()

  def _checkConsistency(self, job=None):
    """FIXME Check consistency of the Whizard2 application, this is called from the `Job` instance.

    :param job: The instance of the job
    :type job: ~ILCDIRAC.Interfaces.API.NewInterface.Job.Job
    :returns: S_OK/S_ERROR
    """
    if not self.version:
      return S_ERROR('No version found!')

    if not self.whizard2SinFile:
      return S_ERROR('No sin file set!')

    if not self.numberOfEvents:
      return S_ERROR('Number of events not set!')

    for process in self._decayProc:
      if process not in self.whizard2SinFile:
        return S_ERROR('Process "%s" not found in sindarin file, please check your inputs' % process)

    if self._jobtype != 'User':
      self._listofoutput.append({"outputFile": "@{OutputFile}", "outputPath": "@{OutputPath}",
                                 "outputDataSE": '@{OutputSE}'})

      if self.eventType != '':
        self.prodparameters['Process'] = self.eventType
      else:
        return S_ERROR('evttype not set, please set event type!')

      self.prodparameters['nbevts'] = self.numberOfEvents

      parsedString = self.whizard2SinFile.replace(" ", "").split()
      sqrtMatches = [x for x in parsedString if x.startswith('sqrts=') and x.endswith('GeV')]
      if not sqrtMatches:
        return S_ERROR('No energy set in sin file, please set "sqrts=...GeV"')
      elif len(sqrtMatches) != 1:
        return S_ERROR('Multiple instances of "sqrts=..GeV" detected, only one can be processed')

      if not self.energy:
        self.prodparameters['Energy'] = sqrtMatches[0].replace("sqrts=", "").replace("GeV", "")
        self.energy = float(self.prodparameters['Energy'])
      else:
        self.whizard2SinFile = re.sub(r"sqrts *= *[0-9.]* *GeV", "sqrts = %s GeV" %
                                      self.energy,
                                      self.whizard2SinFile)
        self.prodparameters['Energy'] = str(self.energy)

      self.prodparameters['SinFile'] = self.whizard2SinFile

      modelMatches = [x for x in parsedString if x.startswith('model=')]
      if not modelMatches:
        return S_ERROR('No model set in sin file, please set "model=..."')
      elif len(modelMatches) != 1:
        return S_ERROR('Multiple instances of "model=..." detected, only one can be processed')
      self.prodparameters['Model'] = modelMatches[0].replace("model=", "")

    return S_OK()

  def _applicationModule(self):
    md1 = self._createModuleDefinition()
    md1.addParameter(Parameter("randomSeed", 0, "int", "", "", False, False,
                               "Random seed for the generator"))
    md1.addParameter(Parameter("debug", False, "bool", "", "", False, False, "debug mode"))
    md1.addParameter(Parameter("whizard2SinFile", '', "string", "", "", False, False, "Whizard2 steering options"))
    md1.addParameter(Parameter('whizard2RawSin', False, 'bool', '', '', False, False, 'use the raw sin file'))
    md1.addParameter(Parameter("decayProc", [], "list", "", "", False, False, "processses to simulate"))
    md1.addParameter(Parameter('integratedProcess', '', 'string', '', '', False, False, 'Integrated Process to use'))
    return md1

  def _applicationModuleValues(self, moduleinstance):
    moduleinstance.setValue("randomSeed", self.randomSeed)
    moduleinstance.setValue("debug", self.debug)
    moduleinstance.setValue("whizard2SinFile", self.whizard2SinFile)
    moduleinstance.setValue("whizard2RawSin", self.whizard2RawSin)
    moduleinstance.setValue("decayProc", self._decayProc)
    moduleinstance.setValue('integratedProcess', self._integratedProcess)

  def _checkWorkflowConsistency(self):
    return self._checkRequiredApp()



#
# Copyright (c) 2009-2022 CERN. All rights nots expressly granted are
# reserved.
#
# This file is part of iLCDirac
# (see ilcdirac.cern.ch, contact: ilcdirac-support@cern.ch).
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# In applying this licence, CERN does not waive the privileges and
# immunities granted to it by virtue of its status as an
# Intergovernmental Organization or submit itself to any jurisdiction.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.
#
"""Tests for the Whizard module."""
from __future__ import absolute_import
import unittest
from mock import mock_open, patch, MagicMock as Mock
import six

from DIRAC import S_OK, S_ERROR
from ILCDIRAC.Interfaces.API.NewInterface.Applications import Whizard
from Tests.Utilities.GeneralUtils import assertEqualsImproved, \
    assertDiracFailsWith, assertDiracSucceeds, assertMockCalls

__RCSID__ = "$Id$"

MODULE_NAME = 'ILCDIRAC.Interfaces.API.NewInterface.Applications.Whizard'
BUILTIN_NAME = 'builtins' if six.PY3 else '__builtin__'

# pylint: disable=protected-access,too-many-public-methods


class TestWhizard(unittest.TestCase):
  """Tests the Whizard class."""

  EXPECTED_PRINTOUT = '&process_input\n process_id = ""\n cm_frame = T\n sqrts = 3000\n luminosity = 0\n polarized_beams = T\n structured_beams = T\n beam_recoil = F\n recoil_conserve_momentum = F\n filename = "whizard"\n directory = ""\n input_file = ""\n input_slha_format = F\n/\n&integration_input\n calls = 1 50000 10 50000 1 1500000\n seed = \n reset_seed_each_process = F\n accuracy_goal = 0\n efficiency_goal = 100\n time_limit_adaptation = 0\n stratified = T\n use_efficiency = F\n weights_power = 0.25\n min_bins = 3\n max_bins = 20\n min_calls_per_bin = 10\n min_calls_per_channel = 0\n write_grids = T\n write_grids_raw = F\n write_grids_file = ""\n write_all_grids = F\n write_all_grids_file = ""\n read_grids = F\n read_grids_raw = F\n read_grids_force = T\n read_grids_file = ""\n generate_phase_space = T\n read_model_file = ""\n write_phase_space_file = ""\n read_phase_space = T\n read_phase_space_file = ""\n phase_space_only = F\n use_equivalences = T\n azimuthal_dependence = F\n write_phase_space_channels_file = ""\n off_shell_lines = 1\n extra_off_shell_lines = 1\n splitting_depth = 1\n exchange_lines = 3\n show_deleted_channels = F\n single_off_shell_decays = T\n double_off_shell_decays = F\n single_off_shell_branchings = T\n double_off_shell_branchings = T\n massive_fsr = T\n threshold_mass = -10\n threshold_mass_t = -10\n default_jet_cut = 10\n default_mass_cut = 4\n default_energy_cut = 10\n default_q_cut = 4\n write_default_cuts_file = ""\n read_cuts_file = ""\n user_cut_mode = 0\n user_weight_mode = 0\n/\n&simulation_input\n n_events = 0\n n_calls = 0\n n_events_warmup = 0\n unweighted = T\n normalize_weight = T\n write_weights = F\n write_weights_file = ""\n safety_factor = 1\n write_events = T\n write_events_format = 20\n write_events_file = ""\n events_per_file = 5000000\n bytes_per_file = 0\n min_file_count = 1\n max_file_count = 999\n write_events_raw = F\n write_events_raw_file = ""\n read_events = F\n read_events_force = T\n read_events_raw_file = ""\n keep_beam_remnants = T\n keep_initials = T\n guess_color_flow = F\n recalculate = F\n fragment = T\n fragmentation_method = 3\n user_fragmentation_mode = 0\n pythia_parameters = "PMAS(25,1)=120.; PMAS(25,2)=0.3605E-02; MSTU(22)=20 ;MSTJ(28)=2 ;PARJ(21)=0.40000;PARJ(41)=0.11000; PARJ(42)=0.52000; PARJ(81)=0.25000; PARJ(82)=1.90000; MSTJ(11)=3; PARJ(54)=-0.03100; PARJ(55)=-0.00200;PARJ(1)=0.08500; PARJ(3)=0.45000; PARJ(4)=0.02500; PARJ(2)=0.31000; PARJ(11)=0.60000; PARJ(12)=0.40000; PARJ(13)=0.72000;PARJ(14)=0.43000; PARJ(15)=0.08000; PARJ(16)=0.08000; PARJ(17)=0.17000; MSTP(3)=1;"\n pythia_processes = ""\n shower = F\n shower_nf = 5\n shower_running_alpha_s = F\n shower_alpha_s = 0.2\n shower_lambda = 0.29\n shower_t_min = 1.0\n shower_md = 0.330\n shower_mu = 0.330\n shower_ms = 0.500\n shower_mc = 1.5\n shower_mb = 4.8\n/\n&diagnostics_input\n chattiness = 4\n catch_signals = T\n time_limit = 0\n warn_empty_channel = F\n screen_events = F\n screen_histograms = F\n screen_diagnostics = F\n show_pythia_banner = T\n show_pythia_initialization = T\n show_pythia_statistics = T\n write_logfile = T\n write_logfile_file = ""\n show_input = T\n show_results = T\n show_phase_space = F\n show_cuts = T\n show_histories = F\n show_history = T\n show_weights = T\n show_event = F\n show_histograms = F\n show_overflow = F\n show_excess = T\n read_analysis_file = ""\n plot_width = 130\n plot_height = 90\n plot_excess = T\n plot_history = T\n plot_grids_channels = ""\n plot_grids_logscale = 10\n slha_rewrite_input = T\n slha_ignore_errors = F\n/\n&parameter_input\n/\n&beam_input\n energy = 0\n angle = 0\n direction = 0 0 0\n vector_polarization = F\n polarization = 0.0 0.0\n particle_code = 0\n particle_name = "e1"\n USER_spectrum_on = T\n USER_spectrum_mode = 11\n ISR_on = T\n ISR_alpha = 0.0072993\n ISR_m_in = 0.000511\n ISR_LLA_order = 3\n ISR_map = T\n EPA_on = F\n EPA_map = T\n EPA_alpha = 0.0072993\n EPA_m_in = 0.000511\n EPA_mX = 4\n EPA_Q_max = 4\n EPA_x0 = 0\n EPA_x1 = 0\n/\n&beam_input\n energy = 0\n angle = 0\n direction = 0 0 0\n vector_polarization = F\n polarization = 0.0 0.0\n particle_code = 0\n particle_name = "E1"\n USER_spectrum_on = T\n USER_spectrum_mode = -11\n ISR_on = T\n ISR_alpha = 0.0072993\n ISR_m_in = 0.000511\n ISR_LLA_order = 3\n ISR_map = T\n EPA_on = F\n EPA_map = T\n EPA_alpha = 0.0072993\n EPA_m_in = 0.000511\n EPA_mX = 4\n EPA_Q_max = 4\n EPA_x0 = 0\n EPA_x1 = 0\n/'

  def setUp(self):
    self.whiz = Whizard()

  def test_getters(self):
    from ILCDIRAC.Core.Utilities.WhizardOptions import WhizardOptions
    expected_pdict = {'OK': True, 'Value': {'integration_input': {},
                                                'simulation_input': {},
                                                'diagnostics_input': {},
                                                'process_input': {},
                                                'beam_input_1': {},
                                                'beam_input_2': {}}}
    pdict = self.whiz.getPDict()
    self.whiz.setEvtType('myevent_test')
    self.assertIsNotNone(pdict)
    self.whiz.setGlobalEvtType('test_myglobalevt')
    self.whiz.setLuminosity(138.312)
    self.whiz.setRandomSeed(9024)
    self.whiz.setParameterDict({'mytestval': True, 'more_entres': 'value',
                                  'something': 9103})
    self.whiz.setGeneratorLevelCuts({'generator': False, 'cuts': 123,
                                       'level': 'OK'})
    assert not self.whiz.willBeCut
    self.whiz.willCut()
    assert not self.whiz.useGridFiles
    self.whiz.usingGridFiles()
    self.whiz.setJobIndex('mytestJobIndexS&P500')
    self.whiz.setModel('mytestSModel')
    self.whiz._wo = WhizardOptions(self.whiz.model)
    with patch('%s.open' % BUILTIN_NAME, mock_open(), create=True) as mo:
      self.whiz.addedtojob = True
      self.whiz.dumpWhizardDotIn('someFile.in')
      self.whiz.addedtojob = False
      mo.assert_called_once_with('someFile.in', 'w')
      assertMockCalls(mo().write, [TestWhizard.EXPECTED_PRINTOUT, '\n'], self)
    assertEqualsImproved((pdict, self.whiz.eventType,
                            self.whiz.globalEventType, self.whiz.luminosity,
                            self.whiz.randomSeed, self.whiz.parameterDict,
                            self.whiz.generatorLevelCuts, self.whiz.willBeCut,
                            self.whiz.useGridFiles, self.whiz.jobIndex,
                            self.whiz.model),
                          (expected_pdict, 'myevent_test', 'test_myglobalevt',
                            138.312, 9024,
                            {'mytestval': True, 'more_entres': 'value',
                              'something': 9103},
                            {'generator': False, 'cuts': 123,
                              'level': 'OK'}, True, True,
                            'mytestJobIndexS&P500', 'mytestSModel'), self)
    assert not self.whiz._errorDict

  def test_setevttype_nostring(self):
    assert not self.whiz._errorDict
    self.whiz.setEvtType(123)
    assertEqualsImproved(self.whiz.eventType, 123, self)
    assert self.whiz._errorDict

  def test_setevttype_already_added(self):
    self.whiz.addedtojob = True
    assertDiracFailsWith(self.whiz.setEvtType('myothertestevent'),
                          'cannot modify', self)

  def test_dumpwhizarddotin(self):
    assert not self.whiz._errorDict
    self.whiz.dumpWhizardDotIn()
    assert self.whiz._errorDict

  def test_setfullparamdict(self):
    self.whiz.setFullParameterDict({'some_entry': True, 'myParamDict': {},
                                      123: 2})
    assert not self.whiz._errorDict
    assertEqualsImproved(self.whiz.fullParameterDict,
                          {'some_entry': True, 'myParamDict': {}, 123: 2},
                          self)

  def test_setfullparamdict_wrongtype(self):
    assert not self.whiz._errorDict
    self.whiz.setFullParameterDict('eihnff')
    assert self.whiz._errorDict

  def test_checkconsistency(self):
    self.whiz.setFullParameterDict({'process_input':
                                      {'process_id': 'myprocessid',
                                        'sqrts': 81},
                                      'simulation_input':
                                      {'n_events': 41}})
    plist_mock = Mock()
    plist_mock.existsProcess.return_value = S_OK(True)
    plist_mock.getCSPath.return_value = 'some_dir/some_cs_path/whizardv874.tgz'
    self.whiz._processlist = plist_mock
    self.whiz._generatormodels.models['sm'] = True
    assertDiracSucceeds(self.whiz._checkConsistency(), self)

  def test_checkconsistency_noenergy(self):
    assertDiracFailsWith(self.whiz._checkConsistency(), 'energy not set',
                          self)

  def test_checkconsistency_nonbevts(self):
    self.whiz.energy = 'blabla'
    assertDiracFailsWith(self.whiz._checkConsistency(),
                          'number of events not set', self)

  def test_checkconsistency_noprocess(self):
    self.whiz.energy = 'blabla'
    self.whiz.numberOfEvents = 13
    assertDiracFailsWith(self.whiz._checkConsistency(), 'process not defined',
                          self)

  def test_checkconsistency_noprocesslist(self):
    self.whiz.energy = 'blabla'
    self.whiz.numberOfEvents = 13
    self.whiz.eventType = 'mytype'
    assertDiracFailsWith(self.whiz._checkConsistency(),
                          'process list was not given', self)

  def test_checkconsistency_checkfields_fails(self):
    self.whiz.setFullParameterDict({'something': 'something_else'})
    assertDiracFailsWith(self.whiz._checkConsistency(),
                          'element something is not in the allowed parameters',
                          self)

  def test_checkconsistency_noevttype(self):
    self.whiz.setFullParameterDict({'process_input':
                                      {'process_id': ''}})
    assertDiracFailsWith(self.whiz._checkConsistency(),
                          'event type not specified', self)

  def test_checkconsistency_energyzero(self):
    self.whiz.setFullParameterDict({'process_input':
                                      {'process_id': '',
                                        'sqrts': 0}})
    self.whiz.eventType = 'some_test_evttype'
    self.whiz.energy = 0
    assertDiracFailsWith(self.whiz._checkConsistency(), 'energy set to 0',
                          self)

  def test_checkconsistency_nbevts_zero(self):
    self.whiz.setFullParameterDict({'process_input':
                                      {'process_id': '',
                                        'sqrts': 0},
                                      'simulation_input':
                                      {'n_events': 0}})
    self.whiz.eventType = 'some_test_evttype'
    self.whiz.energy = 341
    self.whiz.numberOfEvents = 0
    assertDiracFailsWith(self.whiz._checkConsistency(),
                          'number of events set to 0', self)

  def test_checkconsistency_genlevelcut_invalid(self):
    self.whiz.setFullParameterDict({'process_input':
                                      {'process_id': '',
                                        'sqrts': 0},
                                      'simulation_input':
                                      {'n_events': 0}})
    self.whiz.eventType = 'some_test_evttype'
    self.whiz.energy = 341
    self.whiz.numberOfEvents = 1235
    self.whiz._processlist = Mock()
    self.whiz.setGeneratorLevelCuts({'something': True})
    assertDiracFailsWith(self.whiz._checkConsistency(),
                          'type of True is not a list', self)

  def test_checkconsistency_noglobalevt(self):
    self.whiz.setFullParameterDict(
        {'process_input':
         {'process_id': 'process_1 some_test_process more_stuff',
          'sqrts': 0}, 'simulation_input': {'n_events': 0}})
    self.whiz.setEvtType('some_test_evttype other_types ignore_us')
    self.whiz.setGlobalEvtType('')
    self.whiz.energy = 341
    self.whiz.numberOfEvents = 1235
    self.whiz._processlist = Mock()
    self.whiz.setGeneratorLevelCuts({'some_test_process': ['testlist']})
    assertDiracFailsWith(self.whiz._checkConsistency(),
                          'global name must be defined', self)

  def test_checkconsistency_process_not_found(self):
    self.whiz.setFullParameterDict(
        {'process_input':
         {'process_id': 'process_1 some_test_process more_stuff',
          'sqrts': 0}, 'simulation_input': {'n_events': 0}})
    self.whiz.setEvtType('some_test_evttype other_types ignore_us')
    self.whiz.setGlobalEvtType('myGlobalTestEvent')
    self.whiz.energy = 341
    self.whiz.numberOfEvents = 1235
    exists_dict = {'process_1': True, 'some_test_process': False}
    plist_mock = Mock()
    plist_mock.existsProcess.side_effect = lambda process: S_OK(
        exists_dict[process])
    self.whiz._processlist = plist_mock
    self.whiz.setGeneratorLevelCuts({'some_test_process': ['testlist']})
    assertDiracFailsWith(self.whiz._checkConsistency(),
                          'process some_test_process does not exist', self)

  def test_checkconsistency_process_version_incompatible(self):
    self.whiz.setFullParameterDict(
        {'process_input':
         {'process_id': 'process_1 some_test_process more_stuff',
          'sqrts': 0}, 'simulation_input': {'n_events': 0}})
    self.whiz.setEvtType('some_test_evttype other_types ignore_us')
    self.whiz.setGlobalEvtType('myGlobalTestEvent')
    self.whiz.energy = 341
    self.whiz.version = 'v248'
    self.whiz.numberOfEvents = 1235
    exists_dict = {'process_1': True, 'some_test_process': True}
    plist_mock = Mock()
    plist_mock.existsProcess.side_effect = lambda process: S_OK(
        exists_dict[process])
    plist_mock.getCSPath.side_effect = ['some_dir/whizardv248.tgz',
                                         'whizard418.tar.gz']
    self.whiz._processlist = plist_mock
    self.whiz.setGeneratorLevelCuts({'some_test_process': ['testlist']})
    assertDiracFailsWith(self.whiz._checkConsistency(),
                          'not available in the same whizard version', self)

  def test_checkconsistency_unknownmodel(self):
    self.whiz.setFullParameterDict(
        {'process_input':
         {'process_id': 'process_1 some_test_process more_stuff',
          'sqrts': 0}, 'simulation_input': {'n_events': 0}})
    self.whiz.setEvtType('some_test_evttype other_types ignore_us')
    self.whiz.setGlobalEvtType('myGlobalTestEvent')
    self.whiz.energy = 341
    self.whiz.model = 'myObscureModel'
    self.whiz.version = '248'
    self.whiz.numberOfEvents = 1235
    exists_dict = {'process_1': True, 'some_test_process': True,
                    'more_stuff': True}
    plist_mock = Mock()
    plist_mock.existsProcess.side_effect = lambda process: S_OK(
        exists_dict[process])
    plist_mock.getCSPath.return_value = 'some_dir/whizard248.tgz'
    self.whiz._generatormodels.models = {}
    self.whiz._processlist = plist_mock
    self.whiz.setGeneratorLevelCuts({'some_test_process': ['testlist']})
    assertDiracFailsWith(self.whiz._checkConsistency(),
                          'unknown model myobscuremodel', self)

  def test_checkconsistency_illegaloutputfile(self):
    self.whiz.setFullParameterDict({'integration_input': {}})
    self.whiz.setEvtType('some_type')
    self.whiz.globalEventType = ''
    self.whiz.energy = 341
    self.whiz.model = ''
    self.whiz.version = '248'
    self.whiz.outputFile = 'some_dir/more_dirs/output.file'
    self.whiz.numberOfEvents = 1235
    exists_dict = {'some_type': True}
    plist_mock = Mock()
    plist_mock.existsProcess.side_effect = lambda process: S_OK(
        exists_dict[process])
    plist_mock.getCSPath.return_value = 'some_dir/whizard248.tgz'
    self.whiz._generatormodels.models = {}
    self.whiz._processlist = plist_mock
    self.whiz.setGeneratorLevelCuts({'some_test_process': ['testlist']})
    assertDiracFailsWith(self.whiz._checkConsistency(),
                          'is a file name, not a path', self)

  def test_checkconsistency_unknownparam(self):
    self.whiz.setFullParameterDict({})
    self.whiz.setEvtType('some_type')
    self.whiz.setParameterDict({'illegal_key': 'illegal_value'})
    self.whiz.globalEventType = ''
    self.whiz.energy = 341
    self.whiz.model = ''
    self.whiz.version = '248'
    self.whiz.outputFile = ''
    self.whiz.numberOfEvents = 1235
    exists_dict = {'some_type': True}
    plist_mock = Mock()
    plist_mock.existsProcess.side_effect = lambda process: S_OK(
        exists_dict[process])
    plist_mock.getCSPath.return_value = 'some_dir/whizard248.tgz'
    self.whiz._generatormodels.models = {}
    self.whiz._jobtype = 'User'
    self.whiz._processlist = plist_mock
    self.whiz.setGeneratorLevelCuts({'some_test_process': ['testlist']})
    assertDiracFailsWith(self.whiz._checkConsistency(),
                          'unknown parameter illegal_key', self)

  def test_checkconsistency_cornercases(self):
    self.whiz.setFullParameterDict({})
    self.whiz.setEvtType('some_type')
    self.whiz.setParameterDict({'PNAME1': 'mytestpname'})
    self.whiz.globalEventType = ''
    self.whiz.energy = 341
    self.whiz.model = ''
    self.whiz.version = '248'
    self.whiz.outputFile = 'output.file'
    self.whiz.numberOfEvents = 1235
    exists_dict = {'some_type': True}
    plist_mock = Mock()
    plist_mock.existsProcess.side_effect = lambda process: S_OK(
        exists_dict[process])
    plist_mock.getCSPath.return_value = 'some_dir/whizard248.tgz'
    self.whiz._generatormodels.models = {}
    self.whiz._jobtype = 'User'
    self.whiz._processlist = plist_mock
    self.whiz.setGeneratorLevelCuts({'some_test_process': ['testlist']})
    assertDiracSucceeds(self.whiz._checkConsistency(), self)

  def test_applicationmodule(self):
    result = self.whiz._applicationModule()
    self.assertIsNotNone(result)

  def test_applicationmodulevalues(self):
    module_mock = Mock()
    self.whiz._applicationModuleValues(module_mock)
    self.assertTrue(module_mock.setValue.called)
    assertEqualsImproved(len(module_mock.mock_calls), 11, self)

  def test_userjobmodules(self):
    module_mock = Mock()
    assertDiracSucceeds(self.whiz._userjobmodules(module_mock), self)

  def test_prodjobmodules(self):
    module_mock = Mock()
    assertDiracSucceeds(self.whiz._prodjobmodules(module_mock), self)

  def test_userjobmodules_fails(self):
    with patch('%s._setUserJobFinalization' % MODULE_NAME, new=Mock(return_value=S_OK('something'))),\
        patch('%s._setApplicationModuleAndParameters' % MODULE_NAME, new=Mock(return_value=S_ERROR('some_test_err'))):
      assertDiracFailsWith(self.whiz._userjobmodules(None),
                            'userjobmodules failed', self)

  def test_prodjobmodules_fails(self):
    with patch('%s._setApplicationModuleAndParameters' % MODULE_NAME, new=Mock(return_value=S_OK('something'))), \
        patch('%s._setOutputComputeDataList' % MODULE_NAME, new=Mock(return_value=S_ERROR('some_other_test_err'))):
      assertDiracFailsWith(self.whiz._prodjobmodules(None),
                            'prodjobmodules failed', self)



#
# Copyright (c) 2009-2022 CERN. All rights nots expressly granted are
# reserved.
#
# This file is part of iLCDirac
# (see ilcdirac.cern.ch, contact: ilcdirac-support@cern.ch).
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# In applying this licence, CERN does not waive the privileges and
# immunities granted to it by virtue of its status as an
# Intergovernmental Organization or submit itself to any jurisdiction.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.
#
"""Run any application provided by the user. Is used when a specific environment is needed (e.g. ROOT).

:since: Jul 12, 2010
:author: sposs
"""

from __future__ import absolute_import
import os
import re

from DIRAC.Core.Utilities.Subprocess import shellCall
from DIRAC import S_OK, S_ERROR, gLogger

from ILCDIRAC.Workflow.Modules.ModuleBase import ModuleBase

__RCSID__ = '$Id$'
LOG = gLogger.getSubLogger(__name__)


class ApplicationScript(ModuleBase):
  """Default application environment.

  Called GenericApplication in the Interface.
  """

  def __init__(self):
    super(ApplicationScript, self).__init__()
    self.enable = True
    self.script = None
    self.setupScript = ''
    self.arguments = ''
    self.applicationName = 'Application script'
    self.applicationVersion = ''

  def applicationSpecificInputs(self):
    LOG.info("The arguments are %s" % self.arguments)
    if 'ParametricParameters' in self.workflow_commons:
      parametric = ' '
      if isinstance(self.workflow_commons['ParametricParameters'], list):
        parametric = " ".join(self.workflow_commons['ParametricParameters'])
      else:
        parametric = self.workflow_commons['ParametricParameters']
      self.arguments += parametric

    return S_OK()

  def runIt(self):
    """Run the application in a controlled environment."""
    self.result = S_OK()
    if not self.script:
      self.result = S_ERROR('Script undefined.')
    if not self.applicationLog:
      self.applicationLog = '%s.log' % (os.path.basename(self.script))
    if not self.result['OK']:
      LOG.error("Failed with :", self.result['Message'])
      return self.result

    if not self.workflowStatus['OK'] or not self.stepStatus['OK']:
      LOG.verbose('Workflow status = %s, step status = %s' % (self.workflowStatus['OK'], self.stepStatus['OK']))
      return S_OK('ApplicationScript should not proceed as previous step did not end properly')

    cmd = []
    if re.search('.py$', self.script):
      cmd.append('python')
      cmd.append(os.path.basename(self.script))
    else:
      cmd.append("./" + os.path.basename(self.script))
    cmd.append(self.arguments)
    cmd.append(self.extraCLIarguments)

    command = ' '.join(cmd)
    LOG.info('Command = %s' % (command))  # Really print here as this is useful to see

    com = []
    if self.setupScript:
      com.append('source %s' % (self.setupScript))
    cmdSep = 'echo "%s"' % ('=' * 50)
    com.append(cmdSep)
    com.append('env | grep OMP')
    com.append("export OMP_NUM_THREADS=1")
    com.append('echo "Log file from execution of: %s"' % (command))
    com.append(cmdSep)
    com.append('env | sort >> localEnv.log')
    com.append(cmdSep)
    if os.path.exists("./lib"):
      com.append('declare -x LD_LIBRARY_PATH=./lib:$LD_LIBRARY_PATH')
    com.append(command)
    com.append('declare -x appstatus=$?')
    com.append('env | grep OMP')
    com.append('exit $appstatus')
    finalCommand = ';'.join(com)

    self.stdError = ''
    result = shellCall(0, finalCommand, callbackFunction=self.redirectLogOutput, bufferLimit=20971520)
    if not result['OK']:
      LOG.error("Application failed :", result["Message"])
      return S_ERROR('Problem Executing Application')

    resultTuple = result['Value']

    status = resultTuple[0]
    # stdOutput = resultTuple[1]
    # stdError = resultTuple[2]
    LOG.info("Status after %s execution is %s" % (os.path.basename(self.script), str(status)))
    failed = False
    if status != 0:
      LOG.info("%s execution completed with non-zero status:" % os.path.basename(self.script))
      failed = True
    elif len(self.stdError) > 0:
      LOG.info("%s execution completed with application warning:" % os.path.basename(self.script))
      LOG.info(self.stdError)
    else:
      LOG.info("%s execution completed successfully:" % os.path.basename(self.script))

    if failed is True:
      LOG.error("==================================\n StdError:\n")
      LOG.error(self.stdError)
      return S_ERROR('%s Exited With Status %s' % (os.path.basename(self.script), status))

    # Above can't be removed as it is the last notification for user jobs
    self.setApplicationStatus(
        '%s (%s %s) Successful' %
        (os.path.basename(
            self.script),
            self.applicationName,
            self.applicationVersion))
    return S_OK(
        '%s (%s %s) Successful' %
        (os.path.basename(
            self.script),
            self.applicationName,
            self.applicationVersion))



#
# Copyright (c) 2009-2022 CERN. All rights nots expressly granted are
# reserved.
#
# This file is part of iLCDirac
# (see ilcdirac.cern.ch, contact: ilcdirac-support@cern.ch).
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# In applying this licence, CERN does not waive the privileges and
# immunities granted to it by virtue of its status as an
# Intergovernmental Organization or submit itself to any jurisdiction.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.
#
"""Compute the outputdata list for production jobs.

:since:  Jun 30, 2010

:author: sposs
"""

from __future__ import absolute_import
from ILCDIRAC.Workflow.Modules.ModuleBase import ModuleBase

from DIRAC import gLogger, S_OK

__RCSID__ = '$Id$'
LOG = gLogger.getSubLogger(__name__)


class ComputeOutputDataList(ModuleBase):
  """In case the previous module executed properly, add the output to the listoutput.

  This is used in the prduction context to ensure only the files coming from successful applications
  are added to the output list. Otherwise, there is a risk to register corrupted files.
  """

  def __init__(self):
    """Module initialization."""
    super(ComputeOutputDataList, self).__init__()
    self.version = __RCSID__
    self.listoutput = []

  def applicationSpecificInputs(self):
    """Update the workflow_commons dictionary with the current step's output."""
    if 'listoutput' in self.step_commons:
      self.listoutput = self.step_commons['listoutput']

    if 'outputList' in self.workflow_commons:
      self.workflow_commons['outputList'] = self.workflow_commons['outputList'] + self.listoutput
    else:
      self.workflow_commons['outputList'] = self.listoutput

    return S_OK()

  def execute(self):
    """Not much to do..."""
    res = self.resolveInputVariables()
    if not res['OK']:
      LOG.error("Failed to resolve input variables:", res['Message'])
      return res
    return S_OK()



#
# Copyright (c) 2009-2022 CERN. All rights nots expressly granted are
# reserved.
#
# This file is part of iLCDirac
# (see ilcdirac.cern.ch, contact: ilcdirac-support@cern.ch).
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# In applying this licence, CERN does not waive the privileges and
# immunities granted to it by virtue of its status as an
# Intergovernmental Organization or submit itself to any jurisdiction.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.
#
"""This module contains the definition of the different applications that can be used to create jobs.

Example usage:

>>> from ILCDIRAC.Interfaces.API.NewInterface.Applications import *
>>> from ILCDIRAC.Interfaces.API.NewInterface.UserJob import *
>>> from ILCDIRAC.Interfaces.API.DiracILC import DiracILC
>>> dirac = DiracILC()
>>> job = UserJob()
>>> ga = GenericApplication()
>>> ga.setScript("myscript.py")
>>> ga.setArguments("some arguments")
>>> ga.setDependency({"mokka":"v0706P08","marlin":"v0111Prod"})
>>> job.append(ga)
>>> job.submit(dirac)

It's also possible to set all the application's properties in the constructor

>>> ga = GenericApplication({"Script":"myscript.py", "Arguments":"some arguments", \
         "Dependency":{"mokka":"v0706P08","marlin":"v0111Prod"}})

but this is more an expert's functionality.

Running:

>>> help(GenericApplication)

prints out all the available methods.

:author: Stephane Poss
:author: Remi Ete
:author: Ching Bon Lam
"""
from __future__ import absolute_import
__RCSID__ = "$Id$"

__all__ = ['GenericApplication', 'GetSRMFile', '_Root', 'RootScript', 'RootMacro',
           'Whizard', 'Pythia', 'PostGenSelection', 'StdhepCut', 'StdhepCutJava',
           'Mokka', 'SLIC', 'OverlayInput', 'Marlin', 'LCSIM', 'SLICPandora',
           'CheckCollections', 'SLCIOConcatenate', 'SLCIOSplit', 'StdHepSplit',
           'Tomato', 'CheckWNs', 'DDSim', 'Fcc', 'FccSw', 'FccAnalysis', 'Whizard2',
           'KKMC',
           'GaudiApp',
           ]

from ILCDIRAC.Interfaces.API.NewInterface.Applications.GenericApplication import GenericApplication
from ILCDIRAC.Interfaces.API.NewInterface.Applications.GetSRMFile import GetSRMFile
from ILCDIRAC.Interfaces.API.NewInterface.Applications._Root import _Root
from ILCDIRAC.Interfaces.API.NewInterface.Applications.RootScript import RootScript
from ILCDIRAC.Interfaces.API.NewInterface.Applications.RootMacro import RootMacro
from ILCDIRAC.Interfaces.API.NewInterface.Applications.Whizard import Whizard
from ILCDIRAC.Interfaces.API.NewInterface.Applications.Pythia import Pythia
from ILCDIRAC.Interfaces.API.NewInterface.Applications.PostGenSelection import PostGenSelection
from ILCDIRAC.Interfaces.API.NewInterface.Applications.StdhepCut import StdhepCut
from ILCDIRAC.Interfaces.API.NewInterface.Applications.StdhepCutJava import StdhepCutJava
from ILCDIRAC.Interfaces.API.NewInterface.Applications.Mokka import Mokka
from ILCDIRAC.Interfaces.API.NewInterface.Applications.SLIC import SLIC
from ILCDIRAC.Interfaces.API.NewInterface.Applications.OverlayInput import OverlayInput
from ILCDIRAC.Interfaces.API.NewInterface.Applications.Marlin import Marlin
from ILCDIRAC.Interfaces.API.NewInterface.Applications.LCSIM import LCSIM
from ILCDIRAC.Interfaces.API.NewInterface.Applications.SLICPandora import SLICPandora
from ILCDIRAC.Interfaces.API.NewInterface.Applications.CheckCollections import CheckCollections
from ILCDIRAC.Interfaces.API.NewInterface.Applications.SLCIOConcatenate import SLCIOConcatenate
from ILCDIRAC.Interfaces.API.NewInterface.Applications.SLCIOSplit import SLCIOSplit
from ILCDIRAC.Interfaces.API.NewInterface.Applications.StdHepSplit import StdHepSplit
from ILCDIRAC.Interfaces.API.NewInterface.Applications.Tomato import Tomato
from ILCDIRAC.Interfaces.API.NewInterface.Applications.CheckWNs import CheckWNs
from ILCDIRAC.Interfaces.API.NewInterface.Applications.DDSim import DDSim
from ILCDIRAC.Interfaces.API.NewInterface.Applications.Fcc import Fcc
from ILCDIRAC.Interfaces.API.NewInterface.Applications.Fcc import FccSw
from ILCDIRAC.Interfaces.API.NewInterface.Applications.Fcc import FccAnalysis
from ILCDIRAC.Interfaces.API.NewInterface.Applications.Whizard2 import Whizard2
from ILCDIRAC.Interfaces.API.NewInterface.Applications.KKMC import KKMC
from ILCDIRAC.Interfaces.API.NewInterface.Applications.GaudiApp import GaudiApp



#
# Copyright (c) 2009-2022 CERN. All rights nots expressly granted are
# reserved.
#
# This file is part of iLCDirac
# (see ilcdirac.cern.ch, contact: ilcdirac-support@cern.ch).
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# In applying this licence, CERN does not waive the privileges and
# immunities granted to it by virtue of its status as an
# Intergovernmental Organization or submit itself to any jurisdiction.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.
#
"""OutputDataPolicy generates the output data that will be created by a workflow task.

DIRAC assumes an execute() method will exist during usage.
"""

from __future__ import absolute_import
from DIRAC import gLogger
from DIRAC.Interfaces.API.Job import Job
from ILCDIRAC.Core.Utilities.ProductionData import constructProductionLFNs
from ILCDIRAC.Core.Utilities.resolvePathsAndNames import getProdFilenameFromInput

LOG = gLogger.getSubLogger(__name__)
__RCSID__ = "$Id$"

ILDJOBTYPES = ['MCGeneration_ILD',
                'MCSimulation_ILD',
                'MCReconstruction_ILD',
                'MCReconstruction_Overlay_ILD',
                'Split_ILD',
              ]


class OutputDataPolicy(object):
  """The OutputDataPolicy formats output file names.

  This module is called from the TransformationSystem TaskManager
  """

  def __init__(self, paramDict):
    self.paramDict = paramDict

  def execute(self):
    """Execute it."""
    jobDescription = self.paramDict['Job']
    prodID = self.paramDict['TransformationID']
    jobID = self.paramDict['TaskID']
    inputData = self.paramDict['InputData']

    job = Job(jobDescription)
    commons = job._getParameters()  # pylint: disable=protected-access
    code = job.workflow.createCode()
    outputList = []
    for line in code.split("\n"):
      if line.count("listoutput"):
        outputList += eval(line.split("#")[0].split("=")[-1])  # pylint: disable=eval-used

    commons['outputList'] = outputList
    commons['PRODUCTION_ID'] = prodID
    commons['JOB_ID'] = jobID
    if inputData:
      commons['InputData'] = inputData

    result = constructProductionLFNs(commons)
    if not result['OK']:
      LOG.error(result['Message'])
      return result

    if commons['JobType'] in ILDJOBTYPES and commons.get('InputData'):
      inputData = commons['InputData']
      inputData = inputData[0] if isinstance(inputData, list) else inputData
      for index, outputFile in enumerate(result['Value']['ProductionOutputData']):
        outputFileILD = getProdFilenameFromInput(inputData, outputFile, prodID, jobID)
        result['Value']['ProductionOutputData'][index] = outputFileILD
        LOG.debug("Changed output file name from '%s' to '%s' " % (outputFile, outputFileILD))

    return result



#
# Copyright (c) 2009-2022 CERN. All rights nots expressly granted are
# reserved.
#
# This file is part of iLCDirac
# (see ilcdirac.cern.ch, contact: ilcdirac-support@cern.ch).
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# In applying this licence, CERN does not waive the privileges and
# immunities granted to it by virtue of its status as an
# Intergovernmental Organization or submit itself to any jurisdiction.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.
#
"""Get the overlay files.

:since: Jan 27, 2011

:author: sposs
"""

from __future__ import print_function
from __future__ import absolute_import
import glob
import os
import random
import subprocess
import time
from math import ceil, pi, sin, cos

import DIRAC
from DIRAC.DataManagementSystem.Client.DataManager import DataManager
from DIRAC.DataManagementSystem.Utilities.DMSHelpers import DMSHelpers
from DIRAC.Resources.Catalog.FileCatalogClient import FileCatalogClient
from DIRAC.Resources.Storage.StorageElement import StorageElement
from DIRAC.Core.Utilities.ReturnValues import returnSingleResult
from DIRAC.Core.Utilities.Subprocess import shellCall
from DIRAC.ConfigurationSystem.Client.Helpers.Operations import Operations
from DIRAC.ConfigurationSystem.Client.Helpers.Resources import getSites, getSitePath
from DIRAC import S_OK, S_ERROR, gLogger, gConfig

from ILCDIRAC.Workflow.Modules.ModuleBase import ModuleBase
from ILCDIRAC.Core.Utilities.WasteCPU import wasteCPUCycles
from ILCDIRAC.Core.Utilities.OverlayFiles import energyWithLowerCaseUnit
from ILCDIRAC.Core.Utilities.Configuration import getOptionValue
from ILCDIRAC.Core.Utilities.InputFilesUtilities import getNumberOfEvents
from ILCDIRAC.OverlaySystem.Client.OverlaySystemClient import OverlaySystemClient
from six.moves import zip

__RCSID__ = '$Id$'
LOG = gLogger.getSubLogger(__name__)


def allowedBkg(bkg, energy=None, detector=None, detectormodel=None, machine='clic_cdr'):
  """Check is supplied bkg is allowed."""
  #gLogger.info("Those are the arguments: %s, %s, %s, %s, %s" % (bkg, energy, detector, detectormodel, machine) )
  ops = Operations()
  #bkg_allowed = [ 'gghad', 'pairs' ]
  # if not bkg in bkg_allowed:
  #  return S_ERROR( "Bkg not allowed" )
  res = -1
  if energy:
    if detectormodel:
      res = ops.getValue("/Overlay/%s/%s/%s/%s/ProdID" % (machine, energy, detectormodel, bkg), -1)
      if res < 0:
        return S_ERROR("No background to overlay")
    if detector:  # needed for backward compatibility
      res = ops.getValue("/Overlay/%s/%s/%s/%s/ProdID" % (machine, detector, energy, bkg), -1)
      if res < 0:
        return S_ERROR("No background to overlay")
  return S_OK(res)


class OverlayInput (ModuleBase):
  """Download the files for overlay."""

  def __init__(self):
    super(OverlayInput, self).__init__()
    self.enable = True
    self.STEP_NUMBER = ''
    self.applicationName = 'OverlayInput'
    self.curdir = os.getcwd()
    self.applicationLog = ''
    self.printoutflag = ''
    self.prodid = 0
    self.detector = ''  # needed for backward compatibility
    self.detectormodel = ""
    self.energytouse = ''
    self.energy = 0
    self.nbofeventsperfile = 100
    self.lfns = []
    self.nbfilestoget = 0
    self.BkgEvtType = 'gghad'
    self.metaEventType = self.BkgEvtType
    self.BXOverlay = 0
    self.ggtohadint = 3.2
    self.nbsigeventsperfile = 0
    self.nbinputsigfile = 1
    self.NbSigEvtsPerJob = 0
    self.datMan = DataManager()
    self.fcc = FileCatalogClient()
    self.site = DIRAC.siteName()
    self.useEnergyForFileLookup = True
    self.machine = 'clic_cdr'
    self.pathToOverlayFiles = ''
    self.processorName = ''

    self._dmshelpers = DMSHelpers()

  def applicationSpecificInputs(self):

    self.pathToOverlayFiles = self.step_commons.get("pathToOverlayFiles", self.pathToOverlayFiles)

    if 'Detector' in self.step_commons:
      self.detectormodel = self.step_commons['Detector']
    if not self.detectormodel and not self.detector and not self.pathToOverlayFiles:
      return S_ERROR('Detector model not defined')

    if 'Energy' in self.step_commons:
      self.energytouse = self.step_commons['Energy']

    if self.energy:
      self.energytouse = energyWithLowerCaseUnit(self.energy)

    if not self.energytouse and not self.pathToOverlayFiles:
      return S_ERROR("Energy not set anywhere!")

    if 'BXOverlay' in self.step_commons:
      self.BXOverlay = self.step_commons['BXOverlay']
    if not self.BXOverlay:
      return S_ERROR("BXOverlay parameter not defined")

    if 'ggtohadint' in self.step_commons:
      self.ggtohadint = self.step_commons['ggtohadint']

    if 'ProdID' in self.step_commons:
      self.prodid = self.step_commons['ProdID']

    if 'NbSigEvtsPerJob' in self.step_commons:
      self.NbSigEvtsPerJob = self.step_commons['NbSigEvtsPerJob']

    if 'BkgEvtType' in self.step_commons:
      self.BkgEvtType = self.step_commons['BkgEvtType']
    self.metaEventType = self.BkgEvtType

    res = allowedBkg(self.BkgEvtType, self.energytouse, detector=self.detector,
                      detectormodel=self.detectormodel, machine=self.machine)
    if not res['OK']:
      return res
    if res['Value'] < 0 and not self.pathToOverlayFiles:
      return S_ERROR("No suitable ProdID")
    # if 'Site' in self.workflow_commons:
    #  self.site = self.workflow_commons['Site']

    self.useEnergyForFileLookup = self.step_commons.get("useEnergyForFileLookup", self.useEnergyForFileLookup)

    if self.InputData:
      if self.NumberOfEvents:
        self.nbsigeventsperfile = self.NumberOfEvents
      else:
        return S_ERROR("Number of events in the signal file is missing")
      self.nbinputsigfile = len(self.InputData)

    LOG.info("Signal Events Per Job: %d " % self.NbSigEvtsPerJob)
    LOG.info("Background Event Type: %s " % self.BkgEvtType)
    LOG.info("Meta Event Type: %s " % self.metaEventType)
    LOG.info("Background Events per bunch crossing: %3.2f" % self.ggtohadint)
    LOG.info("SignalEventsPerFile: %d " % self.nbsigeventsperfile)

    if not self.NbSigEvtsPerJob and not self.nbsigeventsperfile:
      return S_ERROR("Could not determine the number of signal events per input file")
    return S_OK("Input variables resolved")

  def __getFilesFromFC(self):
    """Get the list of files from the FileCatalog."""
    meta = {}
    if self.energy and self.useEnergyForFileLookup:
      meta['Energy'] = str(int(self.energy))
    meta['EvtType'] = self.BkgEvtType
    meta['Datatype'] = 'SIM'
    if self.detectormodel:
      meta['DetectorModel'] = self.detectormodel
    if self.machine == 'ilc_dbd':
      meta['Machine'] = 'ilc'
    if self.machine == 'clic_cdr':
      meta['Machine'] = 'clic'
    res = None
    if self.detector:
      res = self.ops.getValue("/Overlay/%s/%s/%s/%s/ProdID" % (self.machine, self.detector,
                                                               self.energytouse, self.BkgEvtType), 0)
      self.nbofeventsperfile = self.ops.getValue("/Overlay/%s/%s/%s/%s/NbEvts" % (self.machine,
                                                                                  self.energytouse,
                                                                                  self.detector,
                                                                                  self.BkgEvtType),
                                                 100)

      self.metaEventType = self.ops.getValue("/Overlay/%s/%s/%s/%s/EvtType" % (self.machine,
                                                                                 self.energytouse,
                                                                                 self.detector,
                                                                                 self.BkgEvtType),
                                              self.BkgEvtType)

    else:
      res = self.ops.getValue("/Overlay/%s/%s/%s/%s/ProdID" % (self.machine,
                                                               self.energytouse,
                                                               self.detectormodel,
                                                               self.BkgEvtType),
                              0)
      self.nbofeventsperfile = self.ops.getValue("/Overlay/%s/%s/%s/%s/NbEvts" % (self.machine,
                                                                                  self.energytouse,
                                                                                  self.detectormodel,
                                                                                  self.BkgEvtType),
                                                 100)
      self.metaEventType = self.ops.getValue("/Overlay/%s/%s/%s/%s/EvtType" % (self.machine,
                                                                                 self.energytouse,
                                                                                 self.detectormodel,
                                                                                 self.BkgEvtType),
                                              self.BkgEvtType)

    LOG.info("Number of Events Per BackgroundFile: %d " % self.nbofeventsperfile)

    meta['EvtType'] = self.metaEventType
    meta['ProdID'] = res
    if self.prodid:
      meta['ProdID'] = self.prodid
    LOG.info("Using %s as metadata" % (meta))

    return self.fcc.findFilesByMetadata(meta)

  def __getFilesFromPath(self):
    """Get the list of files from the FileCatalog via the user specified path."""
    meta = {}
    listOfFiles = self.fcc.findFilesByMetadata(meta, self.pathToOverlayFiles)
    if not listOfFiles['OK']:
      return listOfFiles

    # find the number of events from the metadata
    resNE = getNumberOfEvents(listOfFiles['Value'][:1])
    if resNE['OK']:
      self.nbofeventsperfile = int(resNE.get('Value', {}).get('nbevts', self.nbofeventsperfile))
      LOG.info("Number of Events Per BackgroundFile: %d " % self.nbofeventsperfile)
    else:
      LOG.warn('Failed to find the number of events:', resNE['Message'])
      LOG.warn("Using Default number of Events Per BackgroundFile: %d " % self.nbofeventsperfile)

    return listOfFiles

  def __getFilesFromLyon(self, meta):
    """List the files present at Lyon, not used."""
    prodID = meta['ProdID']
    prod = str(prodID).zfill(8)
    energy = meta['Energy']
    bkg = meta["EvtType"]
    detector = meta["DetectorType"]
    path = "/ilc/prod/clic/%s/%s/%s/SIM/%s/" % (energy, bkg, detector, prod)
    comm = ["nsls", "%s" % path]
    res = subprocess.Popen(comm, stdout=subprocess.PIPE).communicate()
    dirlist = res[0].rstrip().split("\n")
    mylist = []
    for mydir in dirlist:
      if mydir.count("dirac_directory"):
        continue
      curdir = path + mydir
      comm2 = ["nsls", curdir]
      res = subprocess.Popen(comm2, stdout=subprocess.PIPE).communicate()
      for oFile in res[0].rstrip().split("\n"):
        if oFile.count("dirac_directory"):
          continue
        mylist.append(path + mydir + "/" + oFile)
    if not mylist:
      return S_ERROR("File list is empty")
    return S_OK(mylist)

  def __getFilesFromCastor(self, meta):
    """Get the available files (list) from the CERN castor storage."""
    prodID = meta['ProdID']
    prod = str(prodID).zfill(8)
    energy = meta['Energy']
    bkg = meta["EvtType"]
    detector = meta["DetectorType"]
    path = "/castor/cern.ch/grid/ilc/prod/%s/%s/%s/%s/SIM/%s/" % (self.machine, energy, bkg, detector, prod)
    comm = ["nsls", "%s" % path]
    res = subprocess.Popen(comm, stdout=subprocess.PIPE).communicate()
    dirlist = res[0].rstrip().split("\n")
    mylist = []
    for mydir in dirlist:
      if mydir.count("dirac_directory"):
        continue
      curdir = path + mydir
      comm2 = ["nsls", curdir]
      res = subprocess.Popen(comm2, stdout=subprocess.PIPE).communicate()
      for oFile in res[0].rstrip().split("\n"):
        if oFile.count("dirac_directory"):
          continue
        mylist.append(path + mydir + "/" + oFile)
    if not mylist:
      return S_ERROR("File list is empty")
    return S_OK(mylist)

  def __getFilesLocaly(self):
    """Download the files."""
    numberofeventstoget = ceil(self.BXOverlay * self.ggtohadint)
    nbfiles = len(self.lfns)
    availableevents = nbfiles * self.nbofeventsperfile
    if availableevents < numberofeventstoget:
      return S_ERROR("Number of %s events available is less than requested" % (self.BkgEvtType))

    if not self.NbSigEvtsPerJob:
      # Compute Nsignal events
      self.NbSigEvtsPerJob = self.nbinputsigfile * self.nbsigeventsperfile
    if not self.NbSigEvtsPerJob:
      return S_ERROR('Could not determine the number of signal events per job')
    LOG.verbose("There are %s signal event" % self.NbSigEvtsPerJob)
    # Now determine how many files are needed to cover all signal events
    totnboffilestoget = int(ceil(self.NbSigEvtsPerJob * numberofeventstoget / self.nbofeventsperfile))

    # Limit ourself to some configuration maximum
    levels = [self.machine, self.energytouse, self.detectormodel, self.BkgEvtType]
    maxNbFilesToGet = getOptionValue(ops=self.ops, basePath="/Overlay", optionName="MaxNbFilesToGet", defaultValue=20,
                                     levels=levels)

    if totnboffilestoget > maxNbFilesToGet:
      totnboffilestoget = maxNbFilesToGet

    self.__disableWatchDog()
    overlaymon = OverlaySystemClient()
    # Now need to check that there are not that many concurrent jobs getting the overlay at the same time
    error_count = 0
    count = 0
    while True:
      if error_count > 10:
        LOG.error('OverlayDB returned too many errors')
        return S_ERROR('Failed to get number of concurrent overlay jobs')

      res = overlaymon.canRun(self.site)
      if not res['OK']:
        error_count += 1
        time.sleep(60)
        continue
      error_count = 0
      # if running < max_concurrent_running:
      if res['Value']:
        break
      else:
        count += 1
        if count > 300:
          return S_ERROR("Waited too long: 5h, so marking job as failed")
        if count % 10 == 0:
          self.setApplicationStatus("Overlay standby number %s" % count)
        time.sleep(60)

    self.__enableWatchDog()

    self.setApplicationStatus('Getting overlay files')

    LOG.info('Will obtain %s files for overlay' % totnboffilestoget)

    os.mkdir("./overlayinput_" + self.metaEventType)
    os.chdir("./overlayinput_" + self.metaEventType)
    filesobtained = []
    usednumbers = []
    fail = False
    fail_count = 0

    max_fail_allowed = self.ops.getValue("/Overlay/MaxFailedAllowed", 20)
    while not len(filesobtained) == totnboffilestoget:
      if fail_count > max_fail_allowed:
        LOG.error('Reached the max fail count')
        fail = True
        break

      fileindex = random.randrange(nbfiles)
      if fileindex in usednumbers:
        LOG.info('File index is already used, re-try')
        continue

      usednumbers.append(fileindex)

      res = S_ERROR('Have not tried anything yet')
      if self.site == 'LCG.CERN.ch':
        res = self.getEOSFile(self.lfns[fileindex])
      elif self.site == 'LCG.IN2P3-CC.fr':
        res = self.getLyonFile(self.lfns[fileindex])
      elif self.site == 'LCG.UKI-LT2-IC-HEP.uk':
        res = self.getImperialFile(self.lfns[fileindex])
      elif self.site == 'LCG.RAL-LCG2.uk':
        res = self.getRALFile(self.lfns[fileindex])
      elif self.site == 'LCG.KEK.jp':
        res = self.getKEKFile(self.lfns[fileindex])

      # In case the specific copying did not work (mostly because the files do
      # not exist locally) try again to get the file via the DataManager
      if not res['OK']:
        res = self.getDataManagerFile(self.lfns[fileindex])

      if not res['OK']:
        LOG.warn('Could not obtain %s' % self.lfns[fileindex])
        fail_count += 1
        # Wait for a random time around 3 minutes
        LOG.verbose("Waste happily some CPU time (on average 3 minutes)")
        resWaste = wasteCPUCycles(60 * random.gauss(3, 0.1))
        if not resWaste['OK']:
          LOG.error("Could not waste as much CPU time as wanted, but whatever!")
        continue

      filesobtained.append(self.lfns[fileindex])

    # If no file could be obtained, need to make sure the job fails
    if len(usednumbers) == nbfiles and not filesobtained:
      LOG.error('Could not get all the files needed')
      fail = True

    # Remove all scripts remaining
    scripts = glob.glob("*.sh")
    for script in scripts:
      os.remove(script)

    # Print the file list
    mylist = os.listdir(os.getcwd())
    LOG.info("List of Overlay files:")
    LOG.info("\n".join(mylist))
    os.chdir(self.curdir)
    res = overlaymon.jobDone(self.site)
    if not res['OK']:
      LOG.error("Could not declare the job as finished getting the files")
    if fail:
      LOG.error("Did not manage to get all files needed, too many errors")
      return S_ERROR("Failed to get files")
    LOG.info('Got all files needed.')
    return S_OK()

  def getCASTORFile(self, lfn):
    """Use xrdcp or rfcp to get the files from castor."""
    prependpath = "/castor/cern.ch/grid"
    if not lfn.count("castor/cern.ch"):
      lfile = prependpath + lfn
    else:
      lfile = lfn
    LOG.info("Getting %s" % lfile)
    #command = "rfcp %s ./"%file

    basename = os.path.basename(lfile)

    if os.path.exists("overlayinput.sh"):
      os.unlink("overlayinput.sh")
    with open("overlayinput.sh", "w") as script:
      script.write('#!/bin/sh \n')
      script.write('###############################\n')
      script.write('# Dynamically generated scrip #\n')
      script.write('###############################\n')
      if 'X509_USER_PROXY' in os.environ:
        script.write("cp %s /tmp/x509up_u%s \n" % (os.environ['X509_USER_PROXY'], os.getuid()))
      script.write('declare -x STAGE_SVCCLASS=ilcdata\n')
      script.write('declare -x STAGE_HOST=castorpublic\n')
      script.write(
          r"xrdcp -s root://castorpublic.cern.ch/%s ./ -OSstagerHost=castorpublic\&svcClass=ilcdata\n" %
          lfile.rstrip())
      #script.write("/usr/bin/rfcp 'rfio://cgenstager.ads.rl.ac.uk:9002?svcClass=ilcTape&path=%s' %s\n"%(lfile,basename))
      script.write("""
if [ ! -s %s ]; then
  echo "Using rfcp instead"
  rfcp %s ./
fi\n""" % (basename, lfile))
      script.write('declare -x appstatus=$?\n')
      script.write('exit $appstatus\n')
    os.chmod("overlayinput.sh", 0o755)
    comm = 'sh -c "./overlayinput.sh"'
    self.result = shellCall(600, comm, callbackFunction=self.redirectLogOutput, bufferLimit=20971520)

    localfile = os.path.basename(lfile)
    if os.path.exists(localfile):
      return S_OK(localfile)

    return S_ERROR("Failed")

  def getEOSFile(self, lfn):
    """Use xrdcp to get the files from EOS."""
    prependpath = "/eos/experiment/clicdp/grid"
    if not lfn.startswith(prependpath):
      lfile = prependpath + lfn
    else:
      lfile = lfn
    LOG.info('Getting File from CERN-EOS', lfile)

    if os.path.exists("overlayinput.sh"):
      os.unlink("overlayinput.sh")
    with open("overlayinput.sh", "w") as script:
      script.write('#!/bin/sh \n')
      script.write('################################\n')
      script.write('# Dynamically generated script #\n')
      script.write('################################\n')
      if 'X509_USER_PROXY' in os.environ:
        script.write("cp %s /tmp/x509up_u%s \n" % (os.environ['X509_USER_PROXY'], os.getuid()))
      script.write("xrdcp -s root://eospublic.cern.ch/%s ./ \n" % lfile.rstrip())
      script.write('declare -x appstatus=$?\n')
      script.write('exit $appstatus\n')
    os.chmod("overlayinput.sh", 0o755)
    comm = 'sh -c "./overlayinput.sh"'
    self.result = shellCall(600, comm, callbackFunction=self.redirectLogOutput, bufferLimit=20971520)

    localfile = os.path.basename(lfile)
    if os.path.exists(localfile):
      LOG.info('Successfully got file from CERN-EOS')
      return S_OK(localfile)

    return S_ERROR("Failed")

  def getLyonFile(self, lfn):
    """Use xrdcp to get the files from Lyon."""
    prependpath = '/pnfs/in2p3.fr/data'
    if not lfn.count('in2p3.fr/data'):
      lfile = prependpath + lfn
    else:
      lfile = lfn
    LOG.info('Getting file from LYON', lfile)
    #command = "rfcp %s ./"%file
    #comm = []
    #comm.append("cp $X509_USER_PROXY /tmp/x509up_u%s"%os.getuid())

    if os.path.exists("overlayinput.sh"):
      os.unlink("overlayinput.sh")
    with open("overlayinput.sh", "w") as script:
      script.write('#!/bin/sh \n')
      script.write('###############################\n')
      script.write('# Dynamically generated scrip #\n')
      script.write('###############################\n')
      script.write("cp %s /tmp/x509up_u%s \n" % (os.environ['X509_USER_PROXY'], os.getuid()))
      script.write(". /afs/in2p3.fr/grid/profiles/lcg_env.sh\n")
      script.write("xrdcp root://ccdcacsn179.in2p3.fr:1094%s ./ -s\n" % lfile.rstrip())
      #script.write("/usr/bin/rfcp 'rfio://cgenstager.ads.rl.ac.uk:9002?svcClass=ilcTape&path=%s' %s\n"%(lfile,basename))
      # script.write("""
# if [ ! -s %s ]; then
#  rfcp %s ./
# fi\n"""%(basename,lfile))
      script.write('declare -x appstatus=$?\n')
      script.write('exit $appstatus\n')
    os.chmod("overlayinput.sh", 0o755)
    comm = 'sh -c "./overlayinput.sh"'
    self.result = shellCall(600, comm, callbackFunction=self.redirectLogOutput, bufferLimit=20971520)

    localfile = os.path.basename(lfile)
    if os.path.exists(localfile):
      LOG.info('Successfully got file from LYON-SE')
      return S_OK(localfile)

    return S_ERROR("Failed")

  def getImperialFile(self, lfn):
    """Use dccp to get the files from the Imperial SE."""
    prependpath = '/pnfs/hep.ph.ic.ac.uk/data'
    if not lfn.count('hep.ph.ic.ac.uk/data'):
      lfile = prependpath + lfn
    else:
      lfile = lfn
    LOG.info('Getting file from IMPERIAL', lfile)
    # Don't check for CPU time as other wise, job can get killed
    self.__disableWatchDog()

    if os.path.exists("overlayinput.sh"):
      os.unlink("overlayinput.sh")
    with open("overlayinput.sh", "w") as script:
      script.write('#!/bin/sh \n')
      script.write('###############################\n')
      script.write('# Dynamically generated scrip #\n')
      script.write('###############################\n')
      script.write("dccp dcap://%s%s ./\n" % (os.environ['VO_ILC_DEFAULT_SE'], lfile.rstrip()))
      #script.write("/usr/bin/rfcp 'rfio://cgenstager.ads.rl.ac.uk:9002?svcClass=ilcTape&path=%s' %s\n"%(lfile,basename))
      # script.write("""
# if [ ! -s %s ]; then
#  rfcp %s ./
# fi\n"""%(basename,lfile))
      script.write('declare -x appstatus=$?\n')
      script.write('exit $appstatus\n')
    os.chmod("overlayinput.sh", 0o755)
    comm = 'sh -c "./overlayinput.sh"'
    self.result = shellCall(600, comm, callbackFunction=self.redirectLogOutput, bufferLimit=20971520)

    localfile = os.path.basename(lfile)
    if os.path.exists(localfile):
      LOG.info('Successfully got file from IMPERIAL')
      return S_OK(localfile)

    return S_ERROR("Failed")

  def getRALFile(self, lfn):
    """Use rfcp to get the files from RAL castor."""
    prependpath = '/castor/ads.rl.ac.uk/prod'
    if not lfn.count('ads.rl.ac.uk/prod'):
      lfile = prependpath + lfn
    else:
      lfile = lfn
    LOG.info('Getting file from RAL', lfile)
    # Don't check for CPU time as other wise, job can get killed
    self.__disableWatchDog()

    #command = "rfcp %s ./"%file
    #comm = []
    #comm.append("cp $X509_USER_PROXY /tmp/x509up_u%s"%os.getuid())
    if 'X509_USER_PROXY' in os.environ:
      comm2 = ["cp", os.environ['X509_USER_PROXY'], "/tmp/x509up_u%s" % os.getuid()]
      res = subprocess.Popen(comm2, stdout=subprocess.PIPE).communicate()
      print(res)
    #comm.append("xrdcp root://ccdcacsn179.in2p3.fr:1094%s ./ -s"%file)
    #command = string.join(comm,";")
    #logfile = file(self.applicationLog,"w")
    os.environ['CNS_HOST'] = 'castorns.ads.rl.ac.uk'
    #comm4= ['declare','-x','CNS_HOST=castorns.ads.rl.ac.uk']
    #res = subprocess.Popen(comm4,stdout=logfile,stderr=subprocess.STDOUT)
    #print res
    os.environ['STAGE_SVCCLASS'] = 'ilcTape'
#     comm5= ['declare','-x','STAGE_SVCCLASS=ilcTape']
#      res = subprocess.call(comm5)
#      print res
    os.environ['STAGE_HOST'] = 'cgenstager.ads.rl.ac.uk'
#      comm6=['declare','-x','STAGE_HOST=cgenstager.ads.rl.ac.uk']
#      res = subprocess.call(comm6)
#      print res
    basename = os.path.basename(lfile)

    if os.path.exists("overlayinput.sh"):
      os.unlink("overlayinput.sh")
    with open("overlayinput.sh", "w") as script:
      script.write('#!/bin/sh \n')
      script.write('###############################\n')
      script.write('# Dynamically generated scrip #\n')
      script.write('###############################\n')
      script.write(
          "/usr/bin/rfcp 'rfio://cgenstager.ads.rl.ac.uk:9002?svcClass=ilcTape&path=%s' %s\n" %
          (lfile, basename))
      script.write('declare -x appstatus=$?\n')
      script.write('exit $appstatus\n')
    os.chmod("overlayinput.sh", 0o755)
    comm = 'sh -c "./overlayinput.sh"'
    self.result = shellCall(600, comm, callbackFunction=self.redirectLogOutput, bufferLimit=20971520)

    localfile = os.path.basename(lfile)
    if os.path.exists(localfile):
      LOG.info('Successfully got file from RAL')
      return S_OK(localfile)

    return S_ERROR("Failed")

  def getKEKFile(self, lfn):
    """Use cp to get the files from kek-se."""
    prependpath = '/grid'
    lfile = prependpath + lfn
    LOG.info('Getting file from KEK', lfile)
    self.__disableWatchDog()

    if os.path.exists("overlayinput.sh"):
      os.unlink("overlayinput.sh")
    with open("overlayinput.sh", "w") as script:
      script.write('#!/bin/sh \n')
      script.write('###############################\n')
      script.write('# Dynamically generated scrip #\n')
      script.write('###############################\n')
      script.write("cp %s ./ -s\n" % lfile.rstrip())
      script.write('declare -x appstatus=$?\n')
      script.write('exit $appstatus\n')

    os.chmod("overlayinput.sh", 0o755)
    comm = 'sh -c "./overlayinput.sh"'
    self.result = shellCall(600, comm, callbackFunction=self.redirectLogOutput, bufferLimit=20971520)

    localfile = os.path.basename(lfile)
    if os.path.exists(localfile):
      LOG.info('Successfully got file from KEK')
      return S_OK(localfile)

    return S_ERROR("Failed")

  def getDataManagerFile(self, lfn):
    """Try to download the file via the DataManager.

    First get the available SEs, then sort them by distance to the Site, then try each SE in turn.

    :param str lfn: the lfn to download
    :return: S_OK/S_ERROR
    """
    self.__disableWatchDog()
    resRepl = returnSingleResult(self.datMan.getActiveReplicas(lfn, getUrl=False))
    # value is a dictionary of {SE:true/false}
    if not resRepl['OK']:
      LOG.error('Failed to get active replicas', '%s: %s' % (lfn, resRepl['Message']))
      return resRepl
    availableAt = [se for se, active in resRepl['Value'].items() if active]
    LOG.info('The file is available at these SEs', '%s: %s' % (lfn, availableAt))
    LOG.info('Running at site', self.site)
    orderedSEs = sorted(availableAt, key=self.__distanceToSE)
    LOG.info('Ordered SEs', orderedSEs)
    for SE in orderedSEs:
      LOG.info('Trying to download from:', SE)
      res = returnSingleResult(self.datMan.getFile(lfn, sourceSE=SE))
      if res['OK']:
        LOG.info('Successfully downloaded')
        return res
      LOG.error('Failed to download from', '%s: %s' % (SE, res['Message']))
    return S_ERROR('Failed to get the file from any sourceSE')

  def execute(self):
    """Run the module, called rom Workflow."""
    self.result = self.resolveInputVariables()
    if not self.result['OK']:
      LOG.error("Failed to resolve input parameters:", self.result['Message'])
      return self.result

    LOG.info("Information after resolveInputVariables:")
    LOG.info("Signal Events Per Job: %d " % self.NbSigEvtsPerJob)
    LOG.info("Background Event Type: %s " % self.BkgEvtType)
    LOG.info("Meta Event Type: %s " % self.metaEventType)
    LOG.info("Background Events per bunch crossing: %3.2f" % self.ggtohadint)
    LOG.info("SignalEventsPerFile: %d " % self.nbsigeventsperfile)

    if not self.applicationLog:
      self.applicationLog = 'Overlay_input.log'
    self.applicationLog = os.path.join(os.getcwd(), self.applicationLog)

    if not self.workflowStatus['OK'] or not self.stepStatus['OK']:
      LOG.verbose('Workflow status = %s, step status = %s' % (self.workflowStatus['OK'], self.stepStatus['OK']))
      return S_OK('OverlayInput should not proceed as previous step did not end properly')
    self.setApplicationStatus('Starting up Overlay')

    if self.pathToOverlayFiles:
      res = self.__getFilesFromPath()
    else:
      res = self.__getFilesFromFC()

    if not res['OK']:
      LOG.error("Failed to get the file list from the catalog:", res["Message"])
      self.setApplicationStatus('OverlayProcessor failed to get file list')
      return res
    else:
      LOG.debug("Found these files: %s" % res)

    self.lfns = res['Value']
    if not self.lfns:
      LOG.error("No Overlay LFNs found")
      self.setApplicationStatus('OverlayProcessor got an empty list')
      return S_ERROR('OverlayProcessor got an empty list')

    res = self.__getFilesLocaly()
    # Now that module is finished,resume CPU time checks
    self.__enableWatchDog()

    if not res['OK']:
      LOG.error("Overlay failed with", res['Message'])
      self.setApplicationStatus('OverlayInput failed to get files locally with message %s' % res['Message'])
      return S_ERROR('OverlayInput failed to get files locally')
    self.setApplicationStatus('OverlayInput finished getting all files successfully')

    # add overlay background information to workflow_commons
    stepNumber = int(self.step_commons['STEP_NUMBER'])
    self.workflow_commons["OI_%i_eventType" % stepNumber] = self.metaEventType
    self.workflow_commons["OI_%i_eventsPerBackgroundFile" % stepNumber] = self.nbofeventsperfile
    self.workflow_commons["OI_%i_processorName" % stepNumber] = self.processorName

    # clear the StorageElementCache to clear sessions
    LOG.info('SEs cached now: %d' % len(StorageElement.seCache.getKeys()))
    StorageElement.seCache.purgeAll()
    LOG.info('SEs cached cleared: %d' % len(StorageElement.seCache.getKeys()))

    return S_OK('OverlayInput finished successfully')

  def __disableWatchDog(self):
    """create the watchdog disable if it does not exists."""
    watchDogFilename = 'DISABLE_WATCHDOG_CPU_WALLCLOCK_CHECK'
    fullPath = os.path.join(self.curdir, watchDogFilename)
    if not os.path.exists(fullPath):
      with open(fullPath, 'w') as checkFile:
        checkFile.write('Dont look at cpu')

  def __enableWatchDog(self):
    """remove the watchdog disable file if it exists."""
    watchDogFilename = 'DISABLE_WATCHDOG_CPU_WALLCLOCK_CHECK'
    fullPath = os.path.join(self.curdir, watchDogFilename)
    if os.path.exists(fullPath):
      os.remove(fullPath)

  def __getSELocation(self, SE):
    """Get the coordinates of the site.

    Using the country code we find all sites in the same country and from that obtain the
    coordinates in latitude and longitude.
    """
    try:
      countryCode = StorageElement(SE).storages[0].getParameters()['Host'].rsplit('.', 1)[-1]
    except Exception as e:
      LOG.error('Failed to instantiate SE', str(e))
      return [-180, -90.0]
    allSites = getSites()
    if not allSites['OK']:
      LOG.error('Could not get allSites', allSites['Message'])
      return [0.0, -90.0]
    allSites = [site for site in allSites['Value'] if site.rsplit('.', 1)[-1] == countryCode]
    LOG.debug('Found these sites in %s' % countryCode, allSites)
    coordinates = [self.__getSiteLocation(site) for site in allSites]
    LOG.debug('Coordinates found ', coordinates)
    averageCoordinates = [sum(y) / len(y) for y in zip(*coordinates)]
    LOG.debug('Avarage Location is', averageCoordinates)
    return averageCoordinates if averageCoordinates else [0.0, 0.0]

  def __getSiteLocation(self, site=None):
    """Return Coordinates of the site."""
    if site is None:
      site = self.site
    sitePath = getSitePath(site)
    if not sitePath['OK']:
      LOG.error('Could not get sitePath for', site + sitePath['Message'])
      return [0.0, 0.0]
    coordinates = gConfig.getValue(os.path.join(sitePath['Value'], 'Coordinates'), ('0.0:0.0')).split(':')
    return [float(coord) for coord in coordinates]

  def __distanceToSE(self, SE):
    """Calculate the distance from the current site to the given SE."""
    seLocation = self.__getSELocation(SE)
    t1 = pi / 2.0 + seLocation[1] * pi / 180.0
    p1 = seLocation[0] * pi / 180.0
    siteLocation = self.__getSiteLocation()
    t2 = pi / 2.0 + siteLocation[1] * pi / 180.0
    p2 = siteLocation[0] * pi / 180.0
    # we do not take the sqrt, because it does not matter for sorting
    distance = 2.0 - 2.0 * (sin(t1) * sin(t2) * cos(p1 - p2) + cos(t1) * cos(t2))
    LOG.debug('Distance found to %s: %s' % (SE, distance))
    return distance



#
# Copyright (c) 2009-2022 CERN. All rights nots expressly granted are
# reserved.
#
# This file is part of iLCDirac
# (see ilcdirac.cern.ch, contact: ilcdirac-support@cern.ch).
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# In applying this licence, CERN does not waive the privileges and
# immunities granted to it by virtue of its status as an
# Intergovernmental Organization or submit itself to any jurisdiction.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.
#
"""Run a GaudiApp."""

from __future__ import absolute_import
import os

from DIRAC.Core.Utilities.Subprocess import shellCall
from DIRAC import S_OK, S_ERROR, gLogger
from ILCDIRAC.Core.Utilities.CombinedSoftwareInstallation import getEnvironmentScript
from ILCDIRAC.Core.Utilities.resolvePathsAndNames import resolveIFpaths, getProdFilename
from ILCDIRAC.Workflow.Modules.ModuleBase import ModuleBase
from ILCDIRAC.Workflow.Utilities.DD4hepMixin import DD4hepMixin


__RCSID__ = '$Id$'
LOG = gLogger.getSubLogger(__name__)


class GaudiAppModule(DD4hepMixin, ModuleBase):
  """Specific Module to run a Gaudi job."""

  def __init__(self):
    super(GaudiAppModule, self).__init__()
    self.enable = True
    self.STEP_NUMBER = ''
    self.result = S_ERROR()
    self.applicationName = 'gaudiapp'
    self.executableName = ''
    self.startFrom = 0
    self.randomSeed = -1
    self.inputFileFlag = ''
    self.outputFileFlag = ''
    self.compactFile = ''
    self.detectorModel = ''
    self.detectorModelFlag = '--GeoSvc.detectors'
    self.eventstring = ['+++ Initializing event']  # FIXME

  def applicationSpecificInputs(self):
    """Resolve all input variables for the module here.

    :return: S_OK()
    """
    if self.WorkflowStartFrom:
      self.startFrom = self.WorkflowStartFrom

    self.randomSeed = self._determineRandomSeed()

    if "IS_PROD" in self.workflow_commons and self.workflow_commons["IS_PROD"]:
      self.OutputFile = getProdFilename(self.OutputFile,
                                        int(self.workflow_commons["PRODUCTION_ID"]),
                                        int(self.workflow_commons["JOB_ID"]),
                                        self.workflow_commons,
                                        )

    LOG.info('Have input files?', self.InputFile)
    LOG.info('Have input data?', self.InputData)
    if not self.InputFile and self.InputData:
      for files in self.InputData:
        self.InputFile.append(files)
        LOG.info('Found input files', self.InputFile)
    return S_OK('Parameters resolved')

  def runIt(self):
    """Execute the following:

    - find the detector model xml, using CS query to obtain the path
    - prepare the steering file and command line parameters
    - run GaudiApp on this steering file and catch the exit status

    :returns: :func:`~DIRAC.Core.Utilities.ReturnValues.S_OK`, :func:`~DIRAC.Core.Utilities.ReturnValues.S_ERROR`
    """
    self.result = S_OK()
    if not self.platform:
      self.result = S_ERROR('No ILC platform selected')
    elif not self.applicationLog:
      self.result = S_ERROR('No Log file provided')
    elif not self.executableName:
      self.result = S_ERROR('No executable name provided')
    if not self.result['OK']:
      LOG.error("Failed to resolve input parameters:", self.result['Message'])
      return self.result

    if not self.workflowStatus['OK'] or not self.stepStatus['OK']:
      LOG.verbose('Workflow status = %s, step status = %s' % (self.workflowStatus['OK'], self.stepStatus['OK']))
      return S_OK('GaudiApp should not proceed as previous step did not end properly')

    # TODO: Setup LD_LIBRARY_PATH for extensions
    res = getEnvironmentScript(self.platform, self.applicationName, self.applicationVersion, self.getEnvScript)
    if not res['OK']:
      LOG.error("Could not obtain the environment script: ", res["Message"])
      return res
    envScriptPath = res["Value"]

    # get the path to the detector model, either local or from the software
    if self.detectorModel:
      resXML = self._getDetectorXML()
      if not resXML['OK']:
        LOG.error("Could not obtain the detector XML file: ", resXML["Message"])
        return resXML
      self.compactFile = resXML['Value']

    if self.InputFile:
      res = resolveIFpaths(self.InputFile)
      if not res['OK']:
        LOG.error("InputFile file not found")
        return res
      self.InputFile = res['Value']

    # if steering file is set try to find it
    if self.SteeringFile:
      self.SteeringFile = os.path.basename(self.SteeringFile)
      if not os.path.exists(self.SteeringFile):
        LOG.error("Missing steering file")
        return S_ERROR("Could not find steering file")

    cmd = []
    if self.InputFile:
      cmd.append(" %s %s" % (self.inputFileFlag, ' '.join(self.InputFile)))

    if self.NumberOfEvents:
      cmd.append("-n %s" % self.NumberOfEvents)

    if self.compactFile:
      cmd.append("%s %s" % (self.detectorModelFlag, self.compactFile))

    cmd.append("%s %s" % (self.outputFileFlag, self.OutputFile))

    scriptName = 'GaudiApp_%s_Run_%s.sh' % (self.applicationVersion, self.STEP_NUMBER)
    if os.path.exists(scriptName):
      os.remove(scriptName)
    script = []
    script.append('#!/bin/bash')
    script.append('source %s' % envScriptPath)
    script.append('echo =========')

    # for user provided libraries
    if os.path.exists("lib"):
      script.append("export LD_LIBRARY_PATH=$PWD/lib:$LD_LIBRARY_PATH")

    script.append('env | sort >> localEnv.log')
    script.append('echo gaudiApp:`which %s`' % self.executableName)
    script.append('echo =========')
    comm = '%(executable)s %(steeringFile)s %(args)s %(extraArgs)s ' % \
        dict(executable=self.executableName,
             args=' '.join(cmd),
             extraArgs=self.extraCLIarguments,
             steeringFile=self.SteeringFile,
             )
    # comm += " -v"

    LOG.info("Command:", comm)
    script.append(comm)
    script.append('declare -x appstatus=$?')
    script.append('exit $appstatus')

    with open(scriptName, 'w') as scriptFile:
      scriptFile.write("\n".join(script))

    if os.path.exists(self.applicationLog):
      os.remove(self.applicationLog)

    os.chmod(scriptName, 0o755)
    comm = 'bash "./%s"' % scriptName
    self.setApplicationStatus('GaudiApp %s step %s' % (self.applicationVersion, self.STEP_NUMBER))
    self.stdError = ''
    self.result = shellCall(0, comm, callbackFunction=self.redirectLogOutput, bufferLimit=20971520)
    resultTuple = self.result['Value']
    if not os.path.exists(self.applicationLog):
      LOG.error("Something went terribly wrong, the log file is not present")
      self.setApplicationStatus('%s failed to produce log file' % (self.applicationName))
      if not self.ignoreapperrors:
        return S_ERROR('%s did not produce the expected log %s' % (self.applicationName, self.applicationLog))
    status = resultTuple[0]

    LOG.info("Status after the application execution is %s" % status)

    return self.finalStatusReport(status)

  def getEnvScript(self, platform, appname, appversion):
    """Not Implemented."""
    return S_ERROR('Not implemented')

  def _determineRandomSeed(self):
    """Determine what the randomSeed should be.

    Depends on production or not.

    .. Note::
      In GaudiApp we use *randomSeed* and not *RandomSeed* as in the other workflow modules

    """
    if self.randomSeed == -1:
      self.randomSeed = self.jobID
    if "IS_PROD" in self.workflow_commons:
      self.randomSeed = int('%s%s' % (self.workflow_commons["PRODUCTION_ID"],
                                      self.workflow_commons["JOB_ID"]))
    return self.randomSeed
